#!/usr/bin/python3
#
#  -------------------------------------------------------------------------
#  MIT License
#  
#  Copyright (c) 2024 Randy
#  
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  -------------------------------------------------------------------------
#
#  This program reads an Nord Stage 4 Program file (ns4p file)
#  and reports its contents in human-readable csv-formatted files
#
#  The main data structures are described at the top of ns4maps.py
#
#  Functions whose names start with the word "interpret" are used
#  by the function interpretParam() to convert a raw integer value
#  to the form similar to the one shown in the Nord Stage 4's display.
#  Each such function has an input called "rawval", which is the
#  un-interpreted integer value of that parameter in the ns4p file.
#
#  Revision history:
#
#  2024-08-19: began development
#  2024-09-06: version 0.0 (first working draft before extensive testing)
#  2024-09-21: version 1.0 (finished resolving issues discovered during extensive testing)
#  2024-09-22: version 1.1 (added ns4decode version to the _master.csv output file)
#  2024-09-23: version 1.2 (added option to process all ns4p files in a folder)
#  2024-09-24: version 1.3 (added a parameter [FX delay tempo MST CLK on/off], 
#                           fixed interpretations of FX mod 1 rate and FX delay tempo 
#                           when MST CLK is on)
#  2024-09-30: version 1.4 (added arpeggiator parameters, contributed by guyd789 from the Nord user forum;
#                           also improved how differences between ns4p file versions
#                           are handled, clarified two parameter-names, 
#                           and reorganized the help text)
#  2024-10-02: version 1.5 (added ability to decode Preset files)
#  2024-10-02:             (treat Live files [ns4l] same as regular Program files,
#                           added error trap when user tries to specify a parameter 
#                           that one of the given files does not have
#                           [example: looking in piano Preset files for synth parameters])
#  2024-10-05: version 1.6 (added user preferences file,
#                           added error trap for index-out-of-bounds 
#                           when user specifies which parameters to decode)
#  2024-10-12: version 1.7 (show rawvalue next to "not decoded",
#                           fixed interpretation of RES/FREQHP when filter type = LP+HP)
#  2024-10-19: version 1.8 (extended the scope of what the --askfirst option controls)
#  2025-01-20: version 1.9 (eliminated dependence on xxd using a new function
#                           contributed by guyd789 from the Nord user forum)
#  2025-02-03: version 1.10 (increased number of decoded piano samples by 7x
#                            thanks to Christian (@florence) and Andreas (@gaaal) 
#                            from the Nord user forum,
#                            and also added synth KB sync, KB hold, and KB hold exclude)
#  2025-02-03: version 1.11 (added synth pan)
#  2025-02-08: version 1.12 (added organ preset on/off,
#                            changed "no morph" -> "none", 
#                            added new options --settings and --file,
#                            replaced help-text with downloadable User Manual)
#  2025-02-14: version 2.0 (added gui option [developed by guyd789],
#                           added --find, --csv, --nocsv, --gui, --nogui,
#                           changed display of split point to be a single note 
#                           instead of a pair of notes,
#                           made a couple of outputs more concise to fit in the gui,
#                           improved some warning/announcement behaviors,
#                           added Live files when decoding everything in a folder,
#                           fixed a bug that occurred when using --exclude with Preset files)
#
from math import log10, floor
from os import popen
from os import path as ospath
from sys import exit

import ns4help   #  holds the text for the various help options
import ns4maps   #  utilities that tell ns4decode where to find each parameter in an ns4p file
import ns4names  #  utilities that tell ns4decode how to convert some integer values to names
import ns4exclude  # utilities to exclude unused paraemters from output (if desired)
import ns4options  # utilities to get options from command-line and from preferences file
import ns4gui    # option for graphic display of outputs

global pianomodels
global samplenames

#  this customizes how much ns4decode writes to the screen while it's running
global verbosity
verbosity = 1

#  different ns4p file format versions exist, 
#  and the interpretations of some parameters are version-dependent
global fileVersion
fileVersion = 0 # will be updated each time an ns4p file is read

global warnAboutOlderFileVersion
global warnAboutNewerFileVersion

#  ns4decode was originally developed only for Program files (fileType = ns4p).
#  The presetPart dictionary is used to adapt it to Preset files (ns4o, ns4n, ns4y)
global presetPart
global fileType
global dontKeep
dontKeep = []

#  used to decide which output files to write, based on input file type
global relevantGroups
relevantGroups = {}
relevantGroups['ns4p'] = ['m', 'o', 'p', 'y'] # Program file has all groups
relevantGroups['ns4o'] = ['m', 'o'] # groups for organ Preset file
relevantGroups['ns4n'] = ['m', 'p'] # groups for piano Preset file
relevantGroups['ns4y'] = ['m', 'y'] # groups for synth Preset file

global valueformat

#  these customize the column-width in the .csv files,
#  which matters when viewing them in a plain-text editor
global columnWidthMin
global columnWidthMax
global columnWidthName
columnWidthMax  = 15  # trigger to call writeAsTwoRows
columnWidthName = 33  # wide enough for most names

notdecoded = 'not decoded'
nomorph = 'none'

# ================================================================
#  Populates this dictionary:
#
#    presetPart[fileType]['exists'][partIndex] 
#           = True if this part of a Program file exists in this type of Preset file
#
#    presetPart[fileType]['offset'][partIndex]
#           = offset (number of bits) of location of this part of the Preset file 
#             relative to its location in the Program file
#
#    presetPart[fileType]['endbefore'][partIndex] 
#           = location of the bit immediately AFTER this part of the Program file
#
def populate_presetPart_info():

    global presetPart

    #  create structure

    presetPart = {}
    presetPart['ns4o'] = {}
    presetPart['ns4n'] = {}
    presetPart['ns4y'] = {}
    
    presetPart['ns4o']['exists'] = []
    presetPart['ns4n']['exists'] = []
    presetPart['ns4y']['exists'] = []
    
    presetPart['ns4o']['offset'] = []
    presetPart['ns4n']['offset'] = []
    presetPart['ns4y']['offset'] = []

    presetPart['ns4o']['endbefore'] = []
    presetPart['ns4n']['endbefore'] = []
    presetPart['ns4y']['endbefore'] = []
    
    numBitsInProgramFile = 868*8

    #  adjustments for organ Preset file

    ft = 'ns4o'
    mainOffset = -45*8

    presetPart[ft]['exists'].append( True )
    presetPart[ft]['offset'].append( 0 )
    presetPart[ft]['endbefore'].append( ns4maps.getLocFromStr( '029-1' ) )

    presetPart[ft]['exists'].append( False ) # exclude the rest of the header
    presetPart[ft]['offset'].append( 0 )
    # endbefore is a little less than the layer on/off ID 
    # because the layer offset is negative in this part of the file
    presetPart[ft]['endbefore'].append( ns4maps.getLocFromStr( '095-1' ) )

    presetPart[ft]['exists'].append( True )
    presetPart[ft]['offset'].append( mainOffset )
    presetPart[ft]['endbefore'].append( ns4maps.getLocFromStr( '113-1' ) )

    presetPart[ft]['exists'].append( False ) # exclude KB zone
    presetPart[ft]['offset'].append( mainOffset )
    presetPart[ft]['endbefore'].append( ns4maps.getLocFromStr( '113-5' ) )

    presetPart[ft]['exists'].append( True )
    presetPart[ft]['offset'].append( mainOffset )
    presetPart[ft]['endbefore'].append( ns4maps.getLocFromStr( '230-1' ) )

    presetPart[ft]['exists'].append( False ) # exclude the piano and synth sections
    presetPart[ft]['offset'].append( mainOffset )
    presetPart[ft]['endbefore'].append( numBitsInProgramFile )

    #  adjustments for piano Preset file

    ft = 'ns4n'
    mainOffset = -180*8

    presetPart[ft]['exists'].append( True )
    presetPart[ft]['offset'].append( 0 )
    presetPart[ft]['endbefore'].append( ns4maps.getLocFromStr( '029-1' ) )

    presetPart[ft]['exists'].append( False ) # exclude the organ section
    presetPart[ft]['offset'].append( 0 )
    # endbefore is a little less than the layer on/off ID 
    # because the layer offset is negative in this part of the file
    presetPart[ft]['endbefore'].append( ns4maps.getLocFromStr( '230-1' ) )

    presetPart[ft]['exists'].append( True )
    presetPart[ft]['offset'].append( mainOffset )
    presetPart[ft]['endbefore'].append( ns4maps.getLocFromStr( '243-1' ) )

    presetPart[ft]['exists'].append( False ) # exclude KB zone
    presetPart[ft]['offset'].append( mainOffset )
    presetPart[ft]['endbefore'].append( ns4maps.getLocFromStr( '243-5' ) )

    presetPart[ft]['exists'].append( True )
    presetPart[ft]['offset'].append( mainOffset )
    presetPart[ft]['endbefore'].append( ns4maps.getLocFromStr( '373-1' ) )

    presetPart[ft]['exists'].append( False ) # exclude the synth section
    presetPart[ft]['offset'].append( mainOffset )
    presetPart[ft]['endbefore'].append( numBitsInProgramFile )

    #  adjustments for synth Preset file

    ft = 'ns4y'
    mainOffset = -327*8

    presetPart[ft]['exists'].append( True )
    presetPart[ft]['offset'].append( 0 )
    presetPart[ft]['endbefore'].append( ns4maps.getLocFromStr( '029-1' ) )

    presetPart[ft]['exists'].append( False ) # exclude the organ and piano sections
    presetPart[ft]['offset'].append( 0 )
    # endbefore is a little less than the layer on/off ID 
    # because the layer offset is negative in this part of the file
    presetPart[ft]['endbefore'].append( ns4maps.getLocFromStr( '377-1' ) )

    presetPart[ft]['exists'].append( True )
    presetPart[ft]['offset'].append( mainOffset )
    presetPart[ft]['endbefore'].append( ns4maps.getLocFromStr( '414-3' ) )

    presetPart[ft]['exists'].append( False ) # exclude KB zone
    presetPart[ft]['offset'].append( mainOffset )
    presetPart[ft]['endbefore'].append( ns4maps.getLocFromStr( '414-6' ) )

    presetPart[ft]['exists'].append( True )
    presetPart[ft]['offset'].append( mainOffset )
    presetPart[ft]['endbefore'].append( ns4maps.getLocFromStr( '445-5' ) )

    presetPart[ft]['exists'].append( False ) # exclude extern CC VALs
    presetPart[ft]['offset'].append( mainOffset )
    presetPart[ft]['endbefore'].append( ns4maps.getLocFromStr( '453-4' ) )

    presetPart[ft]['exists'].append( True )
    presetPart[ft]['offset'].append( mainOffset )
    presetPart[ft]['endbefore'].append( numBitsInProgramFile )

# ================================================================
def shouldKeepThisParam( paramID ):

    global fileType
    global presetPart

    # assumes that paramID = location of bit in layer A
    begloc = ns4maps.getLocFromStr( paramID )

    if fileType in ['ns4o', 'ns4n', 'ns4y']:
        for k in range(len(presetPart[fileType]['endbefore'])):
            if begloc < presetPart[fileType]['endbefore'][k]:
                return presetPart[fileType]['exists'][k]
    return True

# ================================================================
#  given a string of 0s and 1s, this returns the integer defined
#  by interpreting a specified substring as a binary number.
#  The begin-index and end-index of the substring are specified in paramloc
#
#  Assumes that shouldKeepThisParam() has already been checked and is True
#
def getParamValue( bitstring, paramloc ):

    global fileType
    global presetPart

    begloc = paramloc['begloc']
    endloc = paramloc['endloc']
    if fileType in ['ns4o', 'ns4n', 'ns4y']:
        for k in range(len(presetPart[fileType]['endbefore'])):
            if begloc < presetPart[fileType]['endbefore'][k]:
                begloc += presetPart[fileType]['offset'][k]
                endloc += presetPart[fileType]['offset'][k]
                break

    return int( bitstring[begloc:endloc+1], base=2 )

# ================================================================
def enforceBounds( rawval, minval, maxval ):

    rawval = min( [rawval, maxval] )
    rawval = max( [rawval, minval] )
    return rawval

# ================================================================
def interpretFileType( rawval ):

    # https://stackoverflow.com/questions/7396849/convert-binary-to-ascii-and-vice-versa
    result = rawval.to_bytes((rawval.bit_length() + 7) // 8, 'big').decode()
    return result

# ================================================================
def checkFileType( bitstring, filename ):

    global fileType

    # the bits from 009-1 to 012-8 should spell "ns4p"
    begloc = ns4maps.getBitloc(  9, 1 )
    endloc = ns4maps.getBitloc( 12, 8 )
    rawval = int( bitstring[begloc:endloc+1], base=2 )

    fileType = interpretFileType( rawval )

    if fileType == 'ns4l':
        # treat Live files (ns4l) just like regular Program files
        fileType = 'ns4p'

    if fileType not in ['ns4p', 'ns4o', 'ns4n', 'ns4y']:
        print( 'Error: ns4decode only works on ns4p, ns4o, ns4n, and ns4y files, ' )
        if fileType.isalnum():
            print( 'and the header of "' + filename + '" says', "it's a", fileType, 'file.' )
        else:
            print( 'and this does not appear to be any of those types:', filename )
        exit()

# ================================================================
def interpretFileVersion( rawval ):

    global fileType

    if fileType in ['ns4p']:
        val = 256 + rawval
    else:
        val = rawval
    majorVersion = int(val/100)
    minorVersion = val - 100*majorVersion
    return str(majorVersion) + '.' + f"{minorVersion:02d}"

# ================================================================
def fileVersionOlderThan( versionToCompare ):

    global fileType
    global fileVersion
    return fileVersion < versionToCompare[fileType]

# ================================================================
def fileVersionNewerThan( versionToCompare ):

    global fileType
    global fileVersion
    return fileVersion > versionToCompare[fileType]

# ================================================================
#  Thanks to Nord forum user guyd789 for providing this function
#
def file_to_bitstring(filename):
    try:
        # Open the file in binary read mode
        with open(filename, 'rb') as f:
            # Read the content of the file
            file_content = f.read()
            
            # Convert the content to a bitstring (binary representation)
            bitstring = ''.join(format(byte, '08b') for byte in file_content)
            
        return bitstring
    except FileNotFoundError:
        print(f"File {filename} not found.")
        return None

# ================================================================
#  This function is no longer needed
#
#def executeUnixCommand( cmdString ):
#
#    lines = popen( cmdString ).readlines()
#    return lines

# ================================================================
#   this reads an ns4 file and populate the 'rawval' fields in the paramdata structure,
#   using other already-populated fields in the paramdata structure to find
#   where the parameters live in the ns4 file
#
def readFile( filename, paramdata ):

    global dontKeep

    if verbosity >= 2:
        print( 'Reading', filename )

    if not ospath.isfile( filename ):
        print( 'Error: no file called', filename )
        exit()

    if 1:
        # new version from guyd789 that doesn't depend on xxd
        bitstring = file_to_bitstring(filename)
    else:
        # old version (deprecated on 2025-01-20)

        #  (R.S.) I've never had any formal software training, so don't laugh... 
        #  This reads the binary file by executing the unix xxd command
        #  and concatenating the output into a single character-string of 1s and 0s.
        #  I did it this way because it made developing ns4decode easy for me.
        cmdString = 'xxd -b -c 1 "' + filename + '"'
        lines = executeUnixCommand( cmdString )
    
        bitstring = ''
        for n in range(len(lines)):
            line = lines[n]
            words = line.split()
            bitstring += words[1]

    checkFileType( bitstring, filename )

    rawvals = []
    for paramID in paramdata.keys():
        if not shouldKeepThisParam( paramID ):
            dontKeep.append( paramID )
            continue
        layervals = []
        for k in range(paramdata[paramID]['numlayers']):
            paramdata[paramID][k]['rawval'] = getParamValue( bitstring, \
                                                             paramdata[paramID][k] )
    return paramdata

# ================================================================
def oneBasedIndex( rawval ):

    return rawval + 1

# ================================================================
#   this wrapper catches cases where the name-lists in ns4names.py
#   are incomplete, because showing "not decoded" in the output
#   is better than crashing
#
def getName( names, key ):

    if key in names.keys():
        return names[key]

    return notdecoded

# ================================================================
def interpretBank( rawval ):

    global fileType

    if fileType in ['ns4p']:
        return getName( ns4names.Letter, rawval )
    else:
        return str(oneBasedIndex( rawval ))

# ================================================================
#  the NS4 displays the location of a Program in memory using
#  the format X:YY, where X is a letter (A-H) and YY is a
#  2-digit number.  Each digit ranges from 1 to 8.
#  This function constructs YY from an integer that 
#  indexes all of the locations consecutively
#
def interpretLocnInBank( rawval ):

    global fileType

    if fileType in ['ns4p']:
        digit1 = oneBasedIndex( int(rawval / 8) )
        digit2 = oneBasedIndex(     rawval % 8 )
        return str(digit1) + str(digit2)
    else:
        result = str(oneBasedIndex( rawval ))
        if len(result) < 2:
            result = '0'+result
        return result

# ================================================================
# https://stackoverflow.com/questions/29246455/python-setting-decimal-place-range-without-rounding
#
def truncate(f, n):
    return floor(f * (10**n)) / 10**n

# ================================================================
def showRawvalWithNotdecoded( rawval ):

    return str(rawval) + ' (' + notdecoded + ')'

# ================================================================
def interpretProgramTranspose( rawval ):

    if rawval < 8:
        value = rawval
    else:
        value = rawval-16

    return f"{value:2d} semitones"

# ================================================================
def interpretVolume( rawval ):

    # https://www.w3schools.com/python/python_string_formatting.asp

    if rawval > 0:
        return f"{40*log10(rawval/127):.1f} dB"
        
    return 'volume off'

# ================================================================
#  some of the NS4's parameters (like the volume of a layer)
#  are morphable, which means that they can be changed using
#  the wheel, or aftertouch (A.T.), or a control pedal.
#  For most of the morphable paramters, the relationship
#  between the un-morphed rawval and the maximally-morphed value is
#  given by first using
#
#      rawvalMorphed = rawvalUnmorphed + (rawval - 127)
#      rawvalMorphed = enforceBounds( rawvalMorphed, 0, 127 )
#
#  to convert rawval to rawvalMorphed, and then 
#  the displayed value is given by the same function as for the un-morphed value
#  but with rawvalMorphed in place of rawval

# ================================================================
def interpretVolumeMorphed( rawval, dependvals ):

    key = list(dependvals.keys())[0]
    rawvalMorphed = dependvals[key] + (rawval - 127)

    rawvalMorphed = enforceBounds( rawvalMorphed, 0, 127 )

    if rawval != 127:
        return interpretVolume( rawvalMorphed )
    return nomorph

# ================================================================
def interpretPan( rawval ):

    sign = 2*(rawval < 32) - 1

    if rawval == 0:  
        return '0.0'
    elif sign > 0:
        value = rawval/3
        return f"R {value:4.1f}"
    elif sign < 0:
        value = (64-rawval)/3
        return f"L {value:4.1f}"

# ================================================================
def interpretOrganVibChorusType( rawval ):

    if fileVersionOlderThan( ns4maps.fileVersionChanged['OrganVibChorusType'] ):
        # the recipe shown below is only for new versions, 
        # and I don't know what recipe old versions used
        return showRawvalWithNotdecoded( rawval )

    #  this is case I didn't anticipate when I designed ns4decode:
    #  the organ vib/chorus value is not tied to any one layer,
    #  but its interpretation depends on the organ model, 
    #  which is different for different layers. 
    #
    #  instead of doing a bunch of refactoring to make that one unexpected case
    #  look "normal" to the user, this function presents it as what it really is: 
    #  an assignment of one vib/chorus type to each organ model

    valueInBits123 = int(rawval/16)
    valueInBits45  = int((rawval - int(rawval/16)*16)/4) 
    valueInBits67  = rawval - int(rawval/4)*4

    vcType = {}
    vcType['B3']   = getName( ns4names.OrganVibChorusType['B3'],   valueInBits123 )
    vcType['VOX']  = getName( ns4names.OrganVibChorusType['VOX'],  valueInBits45 )
    vcType['FARF'] = getName( ns4names.OrganVibChorusType['FARF'], valueInBits67 )
    vcType['PIPE'] =          ns4names.OrganVibChorusType['PIPE']

    result = ''
    for key in vcType.keys():
        result += key + '->' + vcType[key] + '; '
    result = result[:-2]
    return result

# ================================================================
def interpretKBzones( rawval, dependvals ):

    splits = 4*dependvals['050-2'] + 2*dependvals['050-3'] + dependvals['050-4']

    return ns4names.SplitZones[splits][rawval]

# ================================================================
def interpretDrawbar9( rawval, dependvals ):

    key = list(dependvals.keys())[-1]
    organModel = dependvals[key]

    if organModel == 1:  # VOX
        return getName( ns4names.VoxDrawbar9, rawval )

    return interpretDrawbar( rawval, dependvals )

# ================================================================
def interpretDrawbar( rawval, dependvals ):

    key = list(dependvals.keys())[-1]
    organModel = dependvals[key]

    if organModel == 2:  # FARF
        if rawval < 4:
            return '1+2+3+4'
        else:
            return '5+6+7+8'

    return str(rawval)

# ================================================================
def interpretDrawbar9Morphed( rawval, dependvals ):

    key = list(dependvals.keys())[-1]
    organModel = dependvals[key]

    if organModel == 1:  # VOX
        key = list(dependvals.keys())[0]
        rawvalUnmorphed = dependvals[key]
    
        rawvalMorphed = rawvalUnmorphed+rawval-8
        rawvalMorphed = enforceBounds( rawvalMorphed, 0, 8 )
    
        if rawval == 8:
            return nomorph
        return getName( ns4names.VoxDrawbar9, rawvalMorphed )

    return interpretDrawbarMorphed( rawval, dependvals )

# ================================================================
def interpretDrawbarMorphed( rawval, dependvals ):

    key = list(dependvals.keys())[-1]
    organModel = dependvals[key]

    key = list(dependvals.keys())[0]
    rawvalUnmorphed = dependvals[key]

    rawvalMorphed = rawvalUnmorphed+rawval-8
    rawvalMorphed = enforceBounds( rawvalMorphed, 0, 8 )

    if rawval == 8:
        return nomorph

    if organModel == 2:  # FARF
        if rawvalMorphed < 4:
            return '1+2+3+4'
        else:
            return '5+6+7+8'

    return str(rawvalMorphed)

# ================================================================
def interpretPianoModelSlot( rawval ):

    #  the value in the binary fine specifies the
    #  memory-slot in which the sample is stored,
    #  using a 0-based index that runs over all of the memory-slots
    #  within a given piano-type category.
    #  Any given memory slot may either be empty or be occupied by a sample.

    slotspercat = 20
    n = oneBasedIndex( rawval )

    return f"slot {n:3d}/{slotspercat:2d}"

# ================================================================
def labelPianoModelName( rawval ):

    global pianomodels

    key = str(rawval) 
    # pad with leading zeros, if needed
    while len(key) < 10:
        key = '0' + key
    if key in pianomodels.keys():
        return pianomodels[key]

    return 'name unknown'

# ================================================================
def interpretPianoModelName( rawval ):

    value = rawval
    label = labelPianoModelName( rawval )
    return f"{value:010d} ({label:s})"

# ================================================================
def interprePianoModelVariation( rawval, dependvals ):

    key = list(dependvals.keys())[0]
    pianoModel = dependvals[key]

    if pianoModel == 550238836: # Clavinet D6
        return getName( ns4names.ClavVariation, rawval )
    else:
        return 'none'

# ================================================================
def interpretTimbre( rawval, dependvals ):

    key = list(dependvals.keys())[0]
    pianoType = dependvals[key]

    key = list(dependvals.keys())[1]
    pianoModel = dependvals[key]

    if pianoType in [0,1,4,5]:
        return ns4names.TimbreAcoustic[rawval]
    elif pianoType == 2:
        return ns4names.TimbreElectric[rawval]
    elif pianoType == 3:
        if pianoModel == 550238836: # Clavinet D6
            return ns4names.TimbreClav[rawval]
        else:
            return ns4names.TimbreAcoustic[rawval]
    else:
        return showRawvalWithNotdecoded( rawval )

# ================================================================
def labelSampleName( rawval ):

    global samplenames

    key = str(rawval) 
    # pad with leading zeros, if needed
    while len(key) < 10:
        key = '0' + key
    if key in samplenames.keys():
        return samplenames[key]

    return 'name unknown'

# ================================================================
def interpretSampleName( rawval ):

    value = rawval
    label = labelSampleName( rawval )
    return f"{value:010d} ({label:s})"

# ================================================================
def interpretSampleInfo( rawval ):

    #  the value in the binary fine specifies the
    #  memory-slot in which the sample is stored,
    #  using a 0-based index that runs over all of the memory-slots
    #  starting with the first one that can hold a sample.
    #  Any given memory slot may either be empty or be occupied by a sample.
    #
    #  The samples are organized into categories, each of which
    #  has 100 memory slots, so the sample index divided by 100
    #  (rounded to int) is the category index, and 
    #  the sample index modulo 100 gives the memory slot within that category.
    #  This corresponds to how the samples are listed in the Nord Sound Manager.

    slotspercat = 100
    indexofcat =  int( rawval / slotspercat )
    indexincat =       rawval % slotspercat

    indexincat1 = oneBasedIndex( indexincat )

    if indexofcat in ns4names.SampleCategory.keys():
        catname = ns4names.SampleCategory[indexofcat]
    else:
        catname = 'name unknown'

    return f"slot {indexincat1:3d}/{slotspercat:3d} in cat {catname:s}"

# ================================================================
def interpretPresetID( rawval ):

    result = str(rawval) 
    # pad with leading zeros, if needed
    while len(result) < 10:
        result = '0' + result
    return result

# ================================================================
def interpretVibratoDelay( rawval ):

    if fileVersionOlderThan( ns4maps.fileVersionChanged['VibratoDelay'] ):
        # the formula shown below is only for new versions, 
        # and I don't know what formula old versions used
        return showRawvalWithNotdecoded( rawval )

    value = 0.1 + (3.0 - 0.1) * rawval/29
    return f"{value:3.1f} sec"

# ================================================================
def getSynthTypeFromDependvals( dependvals ):

    return getName( ns4names.SynthType, dependvals['561-3'] )

# ================================================================
def labelSynthCategory( rawval, dependvals ):

    #  pages 28-30 in the NS4 user manual OS V1.0X
    synthtype = getSynthTypeFromDependvals( dependvals )

    if synthtype in ['FM-H', 'FM-I']:
        return ns4names.Letter[rawval]

    if synthtype in ['ANALOG'] and fileVersionOlderThan( ns4maps.fileVersionChanged['SynthTypes'] ):
        return getName( ns4names.SynthCategoryOld[synthtype], rawval )

    if synthtype in ['ANALOG', 'WAVE']:
        return getName( ns4names.SynthCategory[synthtype], rawval )

    return notdecoded

# ================================================================
def interpretSynthType( rawval ):

    value = oneBasedIndex( rawval )
    label = getName( ns4names.SynthType, rawval )
    return f"{value:2d} ({label:s})"
#    return f"{label:s}"

# ================================================================
def interpretSynthCategory( rawval, dependvals ):

    synthtype = getSynthTypeFromDependvals( dependvals )

    value = oneBasedIndex( rawval )
    label = labelSynthCategory( rawval, dependvals )

    if synthtype in ['FM-H', 'FM-I']:
        return label
    else:
        return f"{value:2d} ({label:s})"

# ================================================================
def interpretSynthWave( rawval, dependvals ):

    #  pages 28-30 in the NS4 user manual OS V1.0X
    synthtype = getSynthTypeFromDependvals( dependvals )
    synthcatrawval = dependvals['561-8']
    synthcat = labelSynthCategory( synthcatrawval, dependvals )

    label = notdecoded
    if synthtype == 'ANALOG':
        if fileVersionOlderThan( ns4maps.fileVersionChanged['SynthTypes'] ) and synthcatrawval in [3,6,7]:
            label = getName( ns4names.SynthWaveOld[synthcatrawval], rawval )
        elif synthcat != notdecoded:
            label = getName( ns4names.SynthWave[synthcat], rawval )
    elif synthtype == 'FM-H':
        if rawval > 0:
            return str(rawval)
        else:
            return str(0.5)
    elif synthtype == 'FM-I':
        return str(rawval-12)
    elif synthtype == 'WAVE':
        if synthcat != notdecoded:
            label = getName( ns4names.SynthWave[synthcat], rawval )

    value = oneBasedIndex( rawval )

    return f"{value:2d} ({label:s})"

#    return f"{label:s}"

# ================================================================
def interpretSamplePitchFine( rawval ):

    sign = 2*(rawval < 64) - 1
    
    if sign > 0:
        return f"+{rawval:2d} cent"
    elif sign < 0:
        magnitude = 128-rawval
        return f"-{magnitude:2d} cent"

    return '0'

# ================================================================
def interpretSamplePitchCoarse( rawval ):

    sign = 2*(rawval < 32) - 1
    
    if sign > 0:
        return f"+{rawval:2d} semi"
    elif sign < 0:
        magnitude = 64-rawval
        return f"-{magnitude:2d} semi"

    return '0'

# ================================================================
def interpretAttack( rawval ):

    #  this is an approximate fit to the displayed value:
    #
    #       (0.5 ms) * exp( naturallog(45 sec / 0.5 ms) * (rawval/127)^0.57 )
    #
    #  I decided to use the explicit list of values instead of the fit,
    #  because I had to hand-capture all this data anyway in order to deduce the fit,
    #  and I didn't want to bother getting the rounding and number-of-digits logic to 
    #  perfectly match what the NS4 displays. 

    attack = [ \
        '0.5', '0.6', '0.7', '0.9', '1.1', '1.3', '1.5', \
        '1.8', '2.1', '2.5', '3.0', '3.5', '4.0', '4.7', \
        '5.5', '6.3', '7.3', '8.4', '9.7', \
        '11', '13', '14', '16', '19', '21', '24', '27', \
        '31', '34', '39', '43', '49', '54', '61', '68', \
        '75', '84', '93', \
        '103', '114', '126', '139', '153', '169', '186', \
        '202', '224', '246', '269', '295', '322', '352', \
        '384', '419', '456', '496', '540', '586', '636', \
        '690', '748', '810', '876', '947']

    N = len(attack)

    attack.extend( [ \
        '1.02', '1.10', '1.19', '1.28', '1.38', '1.49', '1.60', \
        '1.72', '1.85', '1.99', '2.13', '2.28', '2.45', '2.62', \
        '2.81', '3.00', '3.21', '3.43', '3.60', '3.91', '4.17', \
        '4.45', '4.74', '5.05', '5.37', '5.72', '6.08', '6.47', \
        '6.87', '7.30', '7.75', '8.22', '8.72', '9.25', '9.80', \
        '10', '11', '12', '12', '13', '14', '15', '15', '16', \
        '17', '18', '19', '20', '21', '22', '24', '25', '26', \
        '27', '29', '30', '32', '34', '35', '37', '39', '41', \
        '43', '45'] )

    if rawval < N:
        return attack[rawval] + ' ms'
    else:
        return attack[rawval] + ' s'

# ================================================================
def interpretRelease( rawval ):

    #  this is an approximate fit to the displayed value:
    #
    #      (3 ms) * exp( naturallog(45 sec / 3 ms) * (rawval/127)^0.63 )
    #
    #  I decided to use the explicit list of values instead of the fit,
    #  because I had to hand-capture all this data anyway in order to deduce the fit,
    #  and I didn't want to bother getting the rounding and number-of-digits logic to 
    #  perfectly match what the NS4 displays. 

    decay = [ \
        '3.0', '3.5', '4.0', '4.6', '5.3', '6.0', '6.9', \
        '7.9', '9.0', \
        '10', '12', '13', '15', '17', '19', '21', '23', \
        '26', '29', '33', '36', '41', '45', '50', '55', \
        '61', '68', '75', '82', '91', \
        '100', '110', '120', '132', '144', '158', '173', \
        '188', '206', '224', '244', '265', '288', '313', \
        '340', '368', '399', '432', '467', '505', '545', \
        '588', '634', '683', '736', '792', '851', '915', \
        '983']

    N = len(decay)

    decay.extend( [ \
        '1.05', '1.13', '1.21', '1.30', '1.39', '1.49', '1.59', \
        '1.70', '1.82', '1.94', '2.07', '2.21', '2.36', '2.51', \
        '2.67', '2.85', '3.03', '3.22', '3.42', '3.64', '3.86', \
        '4.10', '4.35', '4.61', '4.89', '5.18', '5.49', '5.81', \
        '6.15', '6.50', '6.88', '7.27', '7.68', '8.11', '8.57', \
        '9.04', '9.54', \
        '10', '11', '11', '12', '12', '13', '14', '14', '15', \
        '16', '17', '18', '19', '20', '20', '22', '23', '24', \
        '25', '26', '27', '29', '30', '31', '33', '34', '36', \
        '38', '39', '41', '43', '45'] )

    if rawval < N:
        return decay[rawval] + ' ms'
    else:
        return decay[rawval] + ' s'

# ================================================================
def interpretDecay( rawval ):

    result = interpretRelease( rawval )
    if rawval == 127:
        result = 'inf'
    return result

# ================================================================
def interpretFXmod1rate( rawval, dependvals ):

    key = list(dependvals.keys())[-1] # only one key if unmorphed, but two if morphed
    if dependvals[key] == 1:
        #  in this case, the rate is locked to master clock,
        #  which changes the way rawval is interpreted
        numvalues = 15
        idx = int( numvalues * rawval/128 )
        return getName( ns4names.LFOrateFracs, idx )

    return interpretZeroToTen( rawval )

# ================================================================
def interpretFXmod1rateMorphed( rawval, dependvals ):

    key = list(dependvals.keys())[0]
    rawvalMorphed = dependvals[key] + (rawval - 127)

    rawvalMorphed = enforceBounds( rawvalMorphed, 0, 127 )

    if rawval != 127:
        return interpretFXmod1rate( rawvalMorphed, dependvals )
    else:
        return nomorph

# ================================================================
def interpretArpRate( rawval, dependvals ):

    key = list(dependvals.keys())[-1] # only one key if unmorphed, but two if morphed
    if dependvals[key] == 1:
        #  in this case, the rate is locked to master clock,
        #  which changes the way rawval is interpreted
        numvalues = 9
        idx = int( numvalues * rawval/128 )
        return getName( ns4names.ArpRateFracs, idx )

    arpRates = [ \
        '16 bpm', 'N.A. ', '18 bpm', '20 bpm', '24 bpm', '26 bpm', '28 bpm', '30 bpm', \
        '34 bpm', '36 bpm', '38 bpm', '42 bpm', '44 bpm', '46 bpm', '48 bpm', '50 bpm', \
        '54 bpm', '56 bpm', '58 bpm', '60 bpm', '62 bpm', '64 bpm', '66 bpm', '68 bpm', \
        '70 bpm', '72 bpm', '74 bpm', '76 bpm', '78 bpm', 'N.A. ', '80 bpm', '82 bpm', \
        '84 bpm', 'N.A. ', '86 bpm', '88 bpm', '90 bpm', '92 bpm', 'N.A  ', '94 bpm', \
        '96 bpm', '98 bpm', 'N.A. ', '100 bpm', '102 bpm', '104 bpm', '106 bpm', '108 bpm', \
        'N.A. ', '110 bpm', '112 bpm', '114 bpm', '116 bpm', '118 bpm', '120 bpm', '122 bpm', \
        '124 bpm', '126 bpm', '128 bpm', '130 bpm', '132 bpm', '134 bpm', '138 bpm', '140 bpm', \
        '142 bpm', '146 bpm', '148 bpm', '152 bpm', '154 bpm', '158 bpm', '162 bpm', '166 bpm', \
        '170 bpm', '174 bpm', '178 bpm', '182 bpm', '186 bpm', '190 bpm', '196 bpm', '200 bpm', \
        '204 bpm', '210 bpm', '216 bpm', '220 bpm', '226 bpm', '232 bpm', '238 bpm', '244 bpm', \
        '252 bpm', '258 bpm', '266 bpm', '274 bpm', '282 bpm', '290 bpm', '298 bpm', '308 bpm', \
        '318 bpm', '328 bpm', '338 bpm', '350 bpm', '362 bpm', '376 bpm', '392 bpm', '410 bpm', \
        '428 bpm', '450 bpm', '472 bpm', '494 bpm', '520 bpm', '546 bpm', '572 bpm', '602 bpm', \
        '632 bpm', '662 bpm', '696 bpm', '728 bpm', '762 bpm', '798 bpm', '834 bpm', '872 bpm', \
        '910 bpm', '950 bpm', '990 bpm', 'Fast1', 'Fast2', 'Fast3', 'Fast4', 'Fast5' ]

    return arpRates[rawval]

# ================================================================
def interpretArpRateMorphed( rawval, dependvals ):

    key = list(dependvals.keys())[0]
    rawvalUnmorphed = dependvals[key]
    rawvalMorphed = rawvalUnmorphed + (rawval - 127)

    rawvalMorphed = enforceBounds( rawvalMorphed, 0, 127 )

    if rawval != 127:
        return interpretArpRate( rawvalMorphed, dependvals )
    else:
        return nomorph

# ================================================================
def interpretArpPatternLength( rawval ):

    return str(rawval+2)

# ================================================================
#  Given a 32-bit rawval, this function separates it into 16 2-bit rawvals.
#
def separateArpValues( rawval ):

    rawvals = []
    tmp = rawval
    for n in range(16):
        thisval = tmp % 4
        # print( f"{n:2d} {thisval:02b} {tmp:032b}" ) # for testing
        rawvals.append( thisval )
        tmp = tmp//4  # integer division

    return rawvals

# ================================================================
#  Given a 32-bit rawval and a list of 4 characters,
#  this function returns a string of 16 characters 
#  (with spaces between groups of 4, so 19 characters total)
#  given by treating the 32-bit rawval as a list of 16 2-bit values,
#  each of which is an index into the list of 4 characters.
#  
def buildArpString( rawval, chars ):

    rawvals = separateArpValues( rawval )
    result = ''
    for n in range(16):
        if n > 0 and n % 4 == 0:
            result += ' '
        result += chars[rawvals[n]]

    return result

# ================================================================
def interpretArpAccent( rawval ):

    chars = ['.', '>', '.', '.']
    return buildArpString( rawval, chars )

# ================================================================
def interpretArpGate( rawval ):

    chars = ['E', 'q', 'Q', '.']
    return buildArpString( rawval, chars )

# ================================================================
def interpretArpPan( rawval ):

    chars = ['C', 'L', 'R', '.']
    return buildArpString( rawval, chars )

# ================================================================
def interpretLFOrate( rawval, dependvals ):

    key = list(dependvals.keys())[-1] # only one key if unmorphed, but two if morphed
    if dependvals[key] == 1:
        #  in this case, the LFO is locked to master clock,
        #  which changes the way rawval is interpreted
        numvalues = 17
        idx = int( numvalues * rawval/128 )
        return getName( ns4names.LFOrateFracs, idx )

    #  this is a good fit to the displayed value:
    #
    #      (0.03 Hz) * exp( naturallog(523 Hz / 0.03 Hz) * (rawval/127) )
    #
    #  I decided to use the explicit list of values instead of the fit,
    #  because I had to hand-capture all this data anyway in order to deduce the fit,
    #  and I didn't want to bother getting the rounding and number-of-digits logic to 
    #  perfectly match what the NS4 displays. 

    LFOrates = [ \
        '0.03', '0.03', '0.03', '0.04', '0.04', '0.04', '0.05', \
        '0.05', '0.05', '0.06', '0.06', '0.06', '0.07', '0.07', \
        '0.08', '0.09', '0.09', '0.10', '0.11', '0.12', '0.13', \
        '0.14', '0.15', '0.16', '0.17', '0.19', '0.20', '0.22', \
        '0.24', '0.26', '0.28', '0.30', '0.32', '0.35', '0.38', \
        '0.44', '0.47', '0.51', '0.55', '0.60', '0.64', '0.70', \
        '0.75', '0.81', '0.88', '0.95', \
        '1.0', '1.1', '1.2', '1.3', '1.4', '1.5', '1.6', '1.8', \
        '1.9', '2.0', '2.2', '2.4', '2.6', '2.8', '3.0', '3.2', \
        '3.5', '3.8', '4.1', '4.4', '4.8', '5.2', '5.6', '6.0', \
        '6.5', '7.0', '7.6', '8.2', '8.8', '9.5', \
        '10', '11', '12', '13', '14', '15', '16', '18', '19', \
        '21', '22', '24', '26', '28', '30', '33', '35', '38', \
        '41', '45', '48', '52', '56', '61', '65', '71', '76', \
        '82', '89', '96', \
        '104', '112', '121', '131', '141', '153', '165', '178', \
        '192', '208', '224', '242', '262', '283', '305', '330', \
        '356', '385', '415', '449', '484', '523']

    return LFOrates[rawval] + ' Hz'

# ================================================================
def interpretLFOrateMorphed( rawval, dependvals ):

    key = list(dependvals.keys())[0]
    rawvalUnmorphed = dependvals[key]
    rawvalMorphed = rawvalUnmorphed + (rawval - 127)

    rawvalMorphed = enforceBounds( rawvalMorphed, 0, 127 )

    if rawval != 127:
        return interpretLFOrate( rawvalMorphed, dependvals )
    else:
        return nomorph

# ================================================================
def interpretLFOmodAmt( rawval, dependvals ):

    key = list(dependvals.keys())[-1] # only one key if unmorphed, but two if morphed
    if dependvals[key] != 2:
        return interpretZeroToTen( rawval )

    #  this is a good fit to the displayed value, with 1 semi = 100 cent:
    #
    #      (48 semi) * (rawval/127)^(5/2)
    #
    #  I decided to use the explicit list of values instead of the fit,
    #  because I had to hand-capture all this data anyway in order to deduce the fit,
    #  and I didn't want to bother getting the rounding and number-of-digits logic to 
    #  perfectly match what the NS4 displays. 

    LFOmodAmts = [ \
        '0', '0', '0', '0', '0', '1', '2', '3', '4', '6', '8', '10', '13', \
        '16', '19', '23', '27', '31', '36', '41', '50', '53', '59', '66', \
        '74', '82', '91']

    N = len(LFOmodAmts)

    LFOmodAmts.extend( [ \
        '1.0', '1.1', '1.2', '1.3', '1.4', '1.5', '1.7', '1.8', '1.9', '2.0', \
        '2.2', '2.4', '2.5', '2.7', '2.8', '3.0', '3.2', '3.4', '3.6', '3.8', \
        '4.0', '4.2', '4.5', '4.8', '5.0', '5.2', '5.5', '5.8', '6.0', '6.2', \
        '6.5', '6.8', '7.0', '7.3', '7.7', '8.0', '8.3', '8.7', '9.0', '9.3', \
        '9.7', '10.0', '10.6', '11.0', '11.3', '11.7', '12.0', '12.5', '13.0', \
        '13.3', '13.7', '14.0', '14.5', '15.0', '15.5', '16.0', '16.5', '17.0', \
        '17.5', '18.0', '18.5', '19.0', '20.0', '20.5', '21.0', '21.5', '22.0', \
        '22.5', '23.0', '24.0', '24.5', '25.0', '26.0', '26.5', '27.0', '28.0', \
        '28.5', '29.0', '30.0', '30.5', '31.0', '32.0', '33.0', '33.5', '34.0', \
        '35.0', '36.0', '37.0', '37.5', '38.0', '39.0', '40.0', '41.0', '42.0', \
        '42.5', '43.0', '44.0', '45.0', '46.0', '47.0', '48.0'] )

    if rawval < N:
        return LFOmodAmts[rawval] + ' cent'
    else:
        return LFOmodAmts[rawval] + ' semi'

# ================================================================
def interpretLFOmodAmtMorphed( rawval, dependvals ):

    key = list(dependvals.keys())[0]
    rawvalUnmorphed = dependvals[key]
    rawvalMorphed = rawvalUnmorphed + (rawval - 127)

    rawvalMorphed = enforceBounds( rawvalMorphed, 0, 127 )

    if rawval != 127:
        return interpretLFOmodAmt( rawvalMorphed, dependvals )
    else:
        return nomorph

# ================================================================
def interpretFilterFreq( rawval ):

    #  this fit works well except for the last dozen values:
    #
    #       (14 Hz) * exp( naturallog(10 kHz / 14 Hz) * (rawval/114) )

    filterFreqs = [  \
    '14', '15', '15', '16', '17', '18', '19', '21', '22', \
    '23', '24', '26', '27', '29', '31', '33', '35', '37', \
    '39', '41', '44', '46', '49', '52', '55', '58', '62', \
    '65', '69', '73', '78', '82', '87', '92', '98',  \
    '104', '110', '117', '123', '131', '139', '141', '156', \
    '165', '175', '185', '196', '208', '220', '233', '247', \
    '262', '277', '294', '311', '330', '349', '370', '392', \
    '415', '440', '466', '493', '523', '554', '587', '622', \
    '659', '698', '740', '784', '831', '880', '932', '988']

    N = len(filterFreqs)

    filterFreqs.extend( [ \
    '1',   '1.1', '1.2', '1.2', '1.3', '1.4', '1.5', \
    '1.6', '1.7', '1.8', '1.9', '2.0', '2.1', '2.2', \
    '2.3', '2.5', '2.6', '2.8', '3.0', '3.1', '3.3', \
    '3.5', '3.7', '4.0', '4.2', '4.4', '4.7', '5.0', \
    '5.3', '5.6', '5.9', '6.3', '6.6', '7.0', '7.5', \
    '7.9', '8.4', '8.9', '9.4', \
    '10', '11', '11', '12', '13', '14', '16', '18', \
    '19', '21', '24', '26', '29', '32'] )

    if rawval < N:
        return filterFreqs[rawval] + ' Hz'
    else:
        return filterFreqs[rawval] + ' kHz'

# ================================================================
def interpretFilterFreqMorphed( rawval, dependvals ):

    key = list(dependvals.keys())[0]
    rawvalUnmorphed = dependvals[key]
    rawvalMorphed = rawvalUnmorphed + (rawval - 127)

    rawvalMorphed = enforceBounds( rawvalMorphed, 0, 127 )

    if rawval != 127:
        return interpretFilterFreq( rawvalMorphed )
    else:
        return nomorph

# ================================================================
def interpretFilterRes( rawval, dependvals ):

    key = list(dependvals.keys())[-1] # only one key if unmorphed, but two if morphed
    if dependvals[key] in [0,1,2,4,5]:
        return interpretZeroToTen( rawval )

    return interpretFilterFreq( rawval )

# ================================================================
def interpretFilterResMorphed( rawval, dependvals ):

    key = list(dependvals.keys())[0]
    rawvalUnmorphed = dependvals[key]
    rawvalMorphed = rawvalUnmorphed + (rawval - 127)

    rawvalMorphed = enforceBounds( rawvalMorphed, 0, 127 )

    if rawval != 127:
        return interpretFilterRes( rawvalMorphed, dependvals )
    else:
        return nomorph

# ================================================================
def interpretOscEnvAmt( rawval ):

    if rawval < 64:
        value = truncate( rawval/6, 1 )
    else:
        value = truncate( (rawval-128)/6, 1 )

    value = enforceBounds( value, -10, 10 )

    return f"{value:5.1f}"

# ================================================================
def interpretOscEnvAmtMorphed( rawval, dependvals ):

    key = list(dependvals.keys())[0]
    rawvalUnmorphed = dependvals[key]

    if rawval != 120:
        rawvalMorphed = rawvalUnmorphed + rawval-120
    else:
        return nomorph

    return interpretOscEnvAmt( rawvalMorphed )

# ================================================================
def interpretVibratoRate( rawval ):

    # ANOMALY:  this formula gives 5.4 when the NS4 says 5.5
    value = truncate( 2 + (8-2) * rawval/127, 1 )

    return f"{value:2.1f} Hz"

# ================================================================
def interpretVibratoAmount( rawval ):

    value = 10.0 * rawval / 50

    return f"{value:3.1f}"

# ================================================================
#  Converts a seven-bit range to a 0-to-10 range
#
def interpretZeroToTen( rawval ):

    value = truncate( 10*rawval/127, 1 )
    return f"{value:2.1f}"

# ================================================================
def interpretZeroToTenMorphed( rawval, dependvals ):

    key = list(dependvals.keys())[0]
    rawvalMorphed = dependvals[key] + (rawval - 127)

    rawvalMorphed = enforceBounds( rawvalMorphed, 0, 127 )

    if rawval != 127:
        return interpretZeroToTen( rawvalMorphed )
    else:
        return nomorph

# ================================================================
def interpretEQlevel( rawval ):

    maxint = 60
    maxdB = 15.0
    if rawval < 64:
        magnitude = rawval
        signstr = '+'
    else:
        magnitude = 128 - rawval
        signstr = '-'

    value = maxdB * magnitude / maxint

    return f"{signstr:s}{value:.1f} dB"

# ================================================================
def interpretEQmid( rawval, dependvals ):

    key = list(dependvals.keys())[0]
    eqmode = dependvals[key]

    if eqmode in [8,9]:
        # when EQ mode is LP FILTER or HP FILTER, 
        # the displayed value is filter resonance (from 0 to 10)
        if rawval < 64:
            value = rawval
        else:
            value = rawval - 128
        value = 10*(60+value)/120
        value = truncate( value, 1 )
        return f"{value:2.1f} flt res"
    else:
        return interpretEQlevel( rawval )

# ================================================================
def interpretEQfreq( rawval ):

    #  this fit works well for values below 1 kHz:
    #
    #      (200 Hz) * exp( naturallog(999 Hz / 200 Hz) * (rawval/64) )
    #
    #  this fit works well for values above 1 kHz:
    #
    #      (1 kHz) * exp( naturallog(8 kHz / 1 kHz) * (rawval-64)/(127-64) )

    vals = [ \
        200, 205, 210, 215, 221, 226, 232, 238, 244, 250, 257, 263, 270, 277, \
        284, 291, 299, 306, 314, 322, 330, 339, 347, 356, 365, 375, 384, 394, \
        404, 414, 425, 436, 447, 458, 470, 482, 494, 507, 520, 533, 546, 560, \
        575, 589, 604, 620, 635, 652, 668, 685, 703, 721, 739, 758, 777, 797, \
        817, 838, 859, 881, 904, 927, 950, 975, 999]

    N = len(vals)
    
    vals.extend( [1.0, 1.0, 1.1, 1.1, 1.2, 1.2, 1.3, 1.3, 1.4, 1.4, 1.5, 1.5, \
                  1.6, 1.6, 1.7, 1.7, 1.8, 1.8, 1.9, 1.9, 2.0, 2.1, 2.1, 2.2, \
                  2.3, 2.4, 2.4] )

    val = 2.5
    while val < 3.8:
        vals.append( val )
        val += 0.1
    
    # skip 3.8

    vals.extend( [ \
        3.9, 4.0, 4.1, 4.3, 4.4, 4.6, 4.7, 4.9, 5.0, 5.2, \
        5.4, 5.6, 5.8, 5.9, 6.1, 6.3, 6.6, 6.8, 7.0, 7.2, 7.5, 7.7, 8.0] )

    val = vals[rawval]

    if val < 200:
        return f"{val:3.1f} kHz"
    else:
        return f"{int(val):3d} Hz"

# ================================================================
def interpretEQfreqMorphed( rawval, dependvals ):

    key = list(dependvals.keys())[0]
    rawvalMorphed = dependvals[key] + (rawval - 127)

    rawvalMorphed = enforceBounds( rawvalMorphed, 0, 127 )

    if rawval != 127:
        return interpretEQfreq( rawvalMorphed )
    else:
        return nomorph

# ================================================================
def interpretDelayTempo( rawval, dependvals ):

    key = list(dependvals.keys())[-1] # only one key if unmorphed, but two if morphed
    if dependvals[key] == 1:
        #  in this case, the tempo is locked to master clock,
        #  which changes the way rawval is interpreted
        numvalues = 17
        idx = int( numvalues * rawval/128 )
        return getName( ns4names.DelayTempoFracs, idx )

    #  this relationship (not shown here) was apparently hand-crafted 
    #  by the Nord Stage 4 engineers, because it is not given by simple 
    #  mathematically-natural function. It is mostly piecewise-linear 
    #  when expressed in beats per minute.

    bpms =  list(range(40,60,2))
    bpms.extend( list(range(60,73, 3)) )
    bpms.append( 76 )
    bpms.extend( list(range(80,110,2)) )
    bpms.extend( list(range(110,140)) )
    bpms.extend( list(range(140,160,2)) )
    bpms.extend( list(range(2*80,2*100,2*2)) )
    bpms.extend( list(range(2*100,2*120,2*4)) )
    bpms.extend( [2*120, 2*126, 2*132, 2*138, 2*152, 2*160] )
    bpms.extend( list(range(4*84, 4*120, 4*4)) )
    bpms.extend( [4*120, 4*126, 4*132, 4*138, 4*144, 4*152, 4*160, 4*180, 4*200, 4*220, \
        4*240, 4*260, 4*280, 4*300, 4*320, 4*340, 4*360, 4*380, 4*401, 4*440, 4*477, \
        4*500, 4*540, 4*580, 4*621, 4*680, 4*750])
    
    #  dt = (60/bpm) seconds, 
    #  but I didn't want to bother getting the rounding logic to 
    #  perfectly match what the NS4 displays
    dts_ms =  [ 1500, 1420, 1360, 1300, 1250, 1200, 1150, 1110, 1070, 1030, 1000, \
        952, 909, 870, 833, 790, 750, 732, 714, 698, 682, 667, 652, 638, 625, \
        612, 600, 588, 577, 566, 556, 546, 541, 536, 531, 526, 522, 517, 513, \
        508, 504, 500, 496, 492, 488, 484, 480, 476, 472, 469, 465, 462, 458, \
        455, 451, 448, 444, 441, 438, 435, 432, 428, 423, 417, 411, 405, 400, \
        395, 390, 385, 380, 375, 366, 357, 349, 341, 333, 326, 319, 313, 306, \
        300, 289, 278, 268, 259, 250, 238, 227, 217, 198, 188, 179, 170, 163, \
        156, 150, 144, 139, 134, 129, 125, 119, 114, 109, 104, \
        99, 94, 83, 75, 68, 63, 58, 54, 50, 47, 44, 42, 40, 37, \
        34, 31, 30, 28, 26, 24, 22, 20]

    bpm = bpms[rawval]
    dt_ms = dts_ms[rawval]

    if dt_ms > 999:
        dt_str = f"{dt_ms/1000:4.2f} sec"
    else:
        dt_str = f"{dt_ms:d} ms"

    if bpm > 2*160:
        bpm_str =  f"{int(bpm/4):d} x4 bpm"
    elif bpm >= 160:
        bpm_str =  f"{int(bpm/2):d} x2 bpm"
    else:
        bpm_str =  f"{bpm:d} bpm"

    return dt_str + ' (' + bpm_str + ')'

# ================================================================
def interpretDelayTempoMorphed( rawval, dependvals ):

    key = list(dependvals.keys())[0]
    rawvalMorphed = dependvals[key] + (rawval - 127)

    rawvalMorphed = enforceBounds( rawvalMorphed, 0, 127 )

    if rawval != 127:
        return interpretDelayTempo( rawvalMorphed, dependvals )
    else:
        return nomorph

# ================================================================
def interpretExternMorphed( rawval, dependvals ):

    key = list(dependvals.keys())[0]
    rawvalMorphed = dependvals[key] + (rawval - 127)

    rawvalMorphed = enforceBounds( rawvalMorphed, 0, 127 )

    if rawval != 127:
        return str( rawvalMorphed )
    else:
        return nomorph

# ================================================================
def getMorphIDs( listOfIDs ):

    result = []
    for n in range(len(listOfIDs)):
        result.append( ns4maps.shiftBit( listOfIDs[n], 7 ) )
        result.append( ns4maps.shiftBit( listOfIDs[n], 7+8 ) )
        result.append( ns4maps.shiftBit( listOfIDs[n], 7+8+8 ) )

    return result

# ================================================================
def getDrawbarMorphIDs( listOfIDs ):

    result = []
    for n in range(len(listOfIDs)):
        result.append( ns4maps.shiftBit( listOfIDs[n], 4 ) )
        result.append( ns4maps.shiftBit( listOfIDs[n], 4+5 ) )
        result.append( ns4maps.shiftBit( listOfIDs[n], 4+5+5 ) )

    return result

# ================================================================
def getDelayTempoMorphIDs( listOfIDs ):

    result = []
    for n in range(len(listOfIDs)):
        result.append( ns4maps.shiftBit( listOfIDs[n], 7+7 ) )
        result.append( ns4maps.shiftBit( listOfIDs[n], 7+7+8 ) )
        result.append( ns4maps.shiftBit( listOfIDs[n], 7+7+8+8 ) )

    return result

# ================================================================
#  this is the main function that tells ns4decode how to interpret each parameter
#
def interpretParam( paramID, rawval, dependvals ):

    global valueformat

    if valueformat == 'binary':  # useful during development
        return f"{rawval:b}"
    if valueformat == 'decimal':  # useful during development
        return f"{rawval:d}"
    if valueformat == 'hex':  # useful during development
        return f"{rawval:x}"

    if paramID == '009-1':
        return interpretFileType( rawval )

    if paramID in ns4maps.shouldInterpretAs['fileVersion']:
        return interpretFileVersion( rawval )

    elif paramID in ['095-1', '095-4', '230-1', '230-4', '377-1', \
                   '377-6', '402-5']:
        return getName( ns4names.Letter, rawval )

    elif paramID == '013-6':
        return interpretBank( rawval )

    elif paramID == '015-3':
        return interpretLocnInBank( rawval )

    elif paramID in ['050-5', '051-1', '051-5']:
        return getName( ns4names.SplitPoint, rawval )

    elif paramID in ['052-1', '052-3', '052-5']:
        return getName( ns4names.SplitZoneXfade, rawval )

    elif paramID == '052-8':
        return interpretProgramTranspose( rawval )

    elif paramID == '054-6':
        return getName( ns4names.Sections, rawval )

    elif paramID in ns4maps.shouldInterpretAs['presetID']:
        return interpretPresetID( rawval )

    elif paramID == '085-3':
        return getName( ns4names.LayerScene, rawval )

    elif paramID in ns4maps.shouldInterpretAs['MorphableVolume']:
        return interpretVolume( rawval )

    elif paramID in getMorphIDs( ns4maps.shouldInterpretAs['MorphableVolume'] ):
        return interpretVolumeMorphed( rawval, dependvals )

    elif paramID == '108-5':
        return getName( ns4names.RotaryStopPosition, rawval )

    elif paramID == '109-4':
        return getName( ns4names.RotarySpeed, rawval )

    elif paramID == '110-3':
        return interpretOrganVibChorusType( rawval )

    elif paramID in ns4maps.shouldInterpretAs['KBzones']:
        return interpretKBzones( rawval, dependvals )

    elif paramID in ['113-5', '243-5', '414-7']:
        return getName( ns4names.OctaveShift, rawval )

    elif paramID == '114-2':
        return getName( ns4names.OrganModel, rawval )

    # handle this case before the other drawbars because
    # shouldInterpretAs['Drawbar'] includes the 9th drawbar
    # (and that, in turn, is because this is the only place where drawbar 9 
    #  requires special treatment)
    elif paramID in ns4maps.shouldInterpretAs['Drawbar9']: 
        return interpretDrawbar9( rawval, dependvals )

    elif paramID in ns4maps.shouldInterpretAs['Drawbar']: 
        return interpretDrawbar( rawval, dependvals )

    elif paramID in getDrawbarMorphIDs( ns4maps.shouldInterpretAs['Drawbar9'] ):
        return interpretDrawbar9Morphed( rawval, dependvals )

    elif paramID in getDrawbarMorphIDs( ns4maps.shouldInterpretAs['Drawbar'] ):
        return interpretDrawbarMorphed( rawval, dependvals )

    elif paramID == '239-2':
        return getName( ns4names.FXpianoLayerFocus, rawval )

    elif paramID == '244-3':
        return getName( ns4names.PianoType, rawval )

    elif paramID == '244-6':
        return interpretPianoModelSlot( rawval )

    elif paramID == '245-5':
        return interpretPianoModelName( rawval )

    elif paramID == '245-3':
        return interprePianoModelVariation( rawval, dependvals )

    elif paramID == '249-8':
        return getName( ns4names.PianoTouch, rawval )

    elif paramID == '250-7':
        return interpretTimbre( rawval, dependvals )

    elif paramID == '389-8':
        return interpretPan( rawval )

    elif paramID == '402-4':
        return getName( ns4names.FXsynthLayerFocus, rawval )

    elif paramID == '408-1':
        return getName( ns4names.SampleAnalog, rawval )

    elif paramID == '408-7':
        return interpretSampleInfo( rawval )

    elif paramID == '410-3':
        return interpretSampleName( rawval )

    elif paramID == '416-1':
        return getName( ns4names.VibratoOption, rawval )

    elif paramID == '416-8':
        return getName( ns4names.VoicePriority, rawval )

    elif paramID == '424-1':
        return getName( ns4names.ArpMode, rawval )

    elif paramID == '428-4':
        return getName( ns4names.ArpDirection, rawval )

    elif paramID in ns4maps.shouldInterpretAs['ArpRate']:
        return interpretArpRate( rawval, dependvals )

    elif paramID in getMorphIDs( ns4maps.shouldInterpretAs['ArpRate'] ):
        return interpretArpRateMorphed( rawval, dependvals )

    elif paramID == '432-7':
        return interpretArpPatternLength( rawval )

    elif paramID in ns4maps.shouldInterpretAs['ArpAccent']:
        return interpretArpAccent( rawval )

    elif paramID in ns4maps.shouldInterpretAs['ArpGate']:
        return interpretArpGate( rawval )

    elif paramID in ns4maps.shouldInterpretAs['ArpPan']:
        return interpretArpPan( rawval )

    elif paramID == '453-5':
        return interpretVibratoDelay( rawval )

    elif paramID == '561-3':
        return interpretSynthType( rawval )

    elif paramID == '561-8':
        return interpretSynthCategory( rawval, dependvals )

    elif paramID == '562-4':
        return interpretSynthWave( rawval, dependvals )

    elif paramID == '567-1':
        return interpretSamplePitchFine( rawval )

    elif paramID == '567-8':
        return interpretSamplePitchCoarse( rawval )

    elif paramID in ['568-8', '584-4', '599-7']:
        return interpretAttack( rawval )

    elif paramID in ['569-7', '585-3', '600-6']:
        return interpretDecay( rawval )

    elif paramID in ['570-6', '586-2', '601-5']:
        return interpretRelease( rawval )

    elif paramID == '575-6':
        return getName( ns4names.SampleOptions, rawval )

    elif paramID == '575-8':
        return getName( ns4names.LFOtarget, rawval )

    elif paramID == '576-2':
        return getName( ns4names.LFOshape, rawval )

    elif paramID in ns4maps.shouldInterpretAs['LFOrate']:
        return interpretLFOrate( rawval, dependvals )

    elif paramID in getMorphIDs( ns4maps.shouldInterpretAs['LFOrate'] ):
        return interpretLFOrateMorphed( rawval, dependvals )

    elif paramID in ns4maps.shouldInterpretAs['LFOmodAmt']:
        return interpretLFOmodAmt( rawval, dependvals )

    elif paramID in getMorphIDs( ns4maps.shouldInterpretAs['LFOmodAmt'] ):
        return interpretLFOmodAmtMorphed( rawval, dependvals )

    elif paramID == '587-3':
        return getName( ns4names.FilterType, rawval )

    elif paramID in ns4maps.shouldInterpretAs['FilterFreq']:
        return interpretFilterFreq( rawval )

    elif paramID in getMorphIDs( ns4maps.shouldInterpretAs['FilterFreq'] ):
        return interpretFilterFreqMorphed( rawval, dependvals )

    elif paramID in ns4maps.shouldInterpretAs['FilterRes']:
        return interpretFilterRes( rawval, dependvals )

    elif paramID in getMorphIDs( ns4maps.shouldInterpretAs['FilterRes'] ):
        return interpretFilterResMorphed( rawval, dependvals )

    elif paramID in ns4maps.shouldInterpretAs['PstickRng']:
        return getName( ns4names.PstickRng, rawval )

    elif paramID in ns4maps.shouldInterpretAs['OscEnvAmt']:
        return interpretOscEnvAmt( rawval )

    elif paramID in getMorphIDs( ns4maps.shouldInterpretAs['OscEnvAmt'] ):
        return interpretOscEnvAmtMorphed( rawval, dependvals )

    elif paramID == '595-4':
        return getName( ns4names.FilterTrack, rawval )

    elif paramID == '602-6':
        return interpretVibratoRate( rawval )

    elif paramID == '603-5':
        return interpretVibratoAmount( rawval )

    elif paramID in ns4maps.shouldInterpretAs['FXmod1mode']:
        return getName( ns4names.FXmod1mode, rawval )

    elif paramID in ns4maps.shouldInterpretAs['FXmod2mode']:
        return getName( ns4names.FXmod2mode, rawval )

    elif paramID in ns4maps.shouldInterpretAs['EQmid']:
        return interpretEQmid( rawval, dependvals )

    elif paramID in ns4maps.shouldInterpretAs['EQlevel']:
        return interpretEQlevel( rawval )

    elif paramID in ns4maps.shouldInterpretAs['EQfreq']:
        return interpretEQfreq( rawval )

    elif paramID in getMorphIDs( ns4maps.shouldInterpretAs['EQfreq'] ):
        return interpretEQfreqMorphed( rawval, dependvals )

    elif paramID in ns4maps.shouldInterpretAs['FXcompResponse']:
        return getName( ns4names.FXcompResponse, rawval )

    elif paramID in ns4maps.shouldInterpretAs['FXmod1rate']:
        return interpretFXmod1rate( rawval, dependvals )

    elif paramID in getMorphIDs( ns4maps.shouldInterpretAs['FXmod1rate'] ):
        return interpretFXmod1rateMorphed( rawval, dependvals )

    elif paramID in ns4maps.shouldInterpretAs['FXdelayMode']:
        return getName( ns4names.FXdelayMode, rawval )

    elif paramID in ns4maps.shouldInterpretAs['FXdelayEffects']:
        return getName( ns4names.FXdelayEffects, rawval )

    elif paramID in ns4maps.shouldInterpretAs['FXdarkBright']:
        return getName( ns4names.FXreverbDarkBright, rawval )

    elif paramID in ns4maps.shouldInterpretAs['FXdelayFilterType']:
        return getName( ns4names.FXdelayFilterType, rawval )

    elif paramID in ns4maps.shouldInterpretAs['FXreverbType']:
        return getName( ns4names.FXreverbType, rawval )

    elif paramID in ns4maps.shouldInterpretAs['AmpEQmode']:
        return getName( ns4names.AmpEQmode, rawval )

    elif paramID in ns4maps.shouldInterpretAs['DelayTempo']:
        return interpretDelayTempo( rawval, dependvals )

    elif paramID in getDelayTempoMorphIDs( ns4maps.shouldInterpretAs['DelayTempo'] ):
        return interpretDelayTempoMorphed( rawval, dependvals )

    elif paramID in getMorphIDs( ns4maps.shouldInterpretAs['MorphableExtern'] ):
        return interpretExternMorphed( rawval, dependvals )

    elif paramID in ns4maps.shouldInterpretAs['NonMorphableZeroToTen']:
        return interpretZeroToTen( rawval )

    elif paramID in ns4maps.shouldInterpretAs['MorphableZeroToTen']:
        return interpretZeroToTen( rawval )

    elif paramID in getMorphIDs( ns4maps.shouldInterpretAs['MorphableZeroToTen'] ):
        return interpretZeroToTenMorphed( rawval, dependvals )

    elif paramID in ns4maps.shouldInterpretAs['OnOff']:
        return getName( ns4names.OnOff, rawval )

    elif paramID in ns4maps.shouldInterpretAs['IncludeExclude']:
        return getName( ns4names.IncludeExclude, rawval )
    else:
        return str( rawval )

# ================================================================
#  this function makes a map that says which parameters have interpretations that
#  depend the values of which other parameters
#
def populateDependson( paramdata ):

    # the interp of piano variation depends on the piano type and the piano model
    paramdata['245-3']['dependson'] = ['245-5']

    # the interp of timbre depends on the piano type and the piano model
    paramdata['250-7']['dependson'] = ['244-3', '245-5']

    # the interp of main synth-knob 2 depends on the setting of synth-knob 1
    paramdata['561-8']['dependson'] = ['561-3']

    # the interp of main synth-knob 3 depends on the setting of synth-knobs 1 and 2
    paramdata['562-4']['dependson'] = ['561-3', '561-8']

    # the interp of filter resonance depends on the filter type
    for paramID in ns4maps.shouldInterpretAs['FilterRes']:
        paramdata[paramID]['dependson'] = ['587-3']

    # the interp of arp rate depends on mst clk on/off
    for paramID in ns4maps.shouldInterpretAs['ArpRate']:
        paramdata[paramID]['dependson'] = ['428-7']

    # the interp of LFO rate/time depends on whether LFO rate mst clk is on/off
    for paramID in ns4maps.shouldInterpretAs['LFOrate']:
        paramdata[paramID]['dependson'] = ['576-5']

    # the interp of LFO mod amt depends on the LFO target
    for paramID in ns4maps.shouldInterpretAs['LFOmodAmt']:
        paramdata[paramID]['dependson'] = ['575-8']

    # the interp of KB zones depends on which split-points are active
    for paramID in ns4maps.shouldInterpretAs['KBzones']:
        paramdata[paramID]['dependson'] = ['050-2', '050-3', '050-4']

    # the interp of the EQ "MID/RES" knob depends on the EQ mode
    for paramID in ns4maps.shouldInterpretAs['EQmid']:
        paramdata[paramID]['dependson'] = [ns4maps.shiftBit( paramID, 32*8+6 )]

    # the interp of the FX mod 1 rate depends on MST CLK on/off
    for paramID in ns4maps.shouldInterpretAs['FXmod1rate']:
        paramdata[paramID]['dependson'] = [ns4maps.shiftBit( paramID, -1 )]

    # the interp of the FX delay tempo depends on MST CLK on/off
    for paramID in ns4maps.shouldInterpretAs['DelayTempo']:
        paramdata[paramID]['dependson'] = [ns4maps.shiftBit( paramID, -1 )]

    # the interp of organ drawbar values depends on the organ model
    for paramID in ns4maps.shouldInterpretAs['Drawbar']:
        paramdata[paramID]['dependson'] = ['114-2']

    # the interp of morph values depends on the non-morphed value

    for paramID in ns4maps.shouldInterpretAs['MorphableVolume']:
        for key in getMorphIDs( [paramID] ):
            paramdata[key]['dependson'] = [paramID]

    for paramID in ns4maps.shouldInterpretAs['EQfreq']:
        for key in getMorphIDs( [paramID] ):
            paramdata[key]['dependson'] = [paramID]

    for paramID in ns4maps.shouldInterpretAs['LFOrate']:
        for key in getMorphIDs( [paramID] ):
            paramdata[key]['dependson'] = [paramID]
            paramdata[key]['dependson'].extend( paramdata[paramID]['dependson'] )

    for paramID in ns4maps.shouldInterpretAs['ArpRate']:
        for key in getMorphIDs( [paramID] ):
            paramdata[key]['dependson'] = [paramID]
            paramdata[key]['dependson'].extend( paramdata[paramID]['dependson'] )

    for paramID in ns4maps.shouldInterpretAs['LFOmodAmt']:
        for key in getMorphIDs( [paramID] ):
            paramdata[key]['dependson'] = [paramID]
            paramdata[key]['dependson'].extend( paramdata[paramID]['dependson'] )

    for paramID in ns4maps.shouldInterpretAs['MorphableExtern']:
        for key in getMorphIDs( [paramID] ):
            paramdata[key]['dependson'] = [paramID]

    for paramID in ns4maps.shouldInterpretAs['MorphableZeroToTen']:
        for key in getMorphIDs( [paramID] ):
            paramdata[key]['dependson'] = [paramID]

    for paramID in ns4maps.shouldInterpretAs['OscEnvAmt']:
        for key in getMorphIDs( [paramID] ):
            paramdata[key]['dependson'] = [paramID]

    for paramID in ns4maps.shouldInterpretAs['FilterFreq']:
        for key in getMorphIDs( [paramID] ):
            paramdata[key]['dependson'] = [paramID]

    for paramID in ns4maps.shouldInterpretAs['FilterRes']:
        for key in getMorphIDs( [paramID] ):
            paramdata[key]['dependson'] = [paramID]
            paramdata[key]['dependson'].extend( paramdata[paramID]['dependson'] )

    for paramID in ns4maps.shouldInterpretAs['Drawbar']:
        for key in getDrawbarMorphIDs( [paramID] ):
            paramdata[key]['dependson'] = [paramID]
            paramdata[key]['dependson'].extend( paramdata[paramID]['dependson'] )

    for paramID in ns4maps.shouldInterpretAs['FXmod1rate']:
        for key in getMorphIDs( [paramID] ):
            paramdata[key]['dependson'] = [paramID]
            paramdata[key]['dependson'].extend( paramdata[paramID]['dependson'] )

    for paramID in ns4maps.shouldInterpretAs['DelayTempo']:
        for key in getDelayTempoMorphIDs( [paramID] ):
            paramdata[key]['dependson'] = [paramID]
            paramdata[key]['dependson'].extend( paramdata[paramID]['dependson'] )

    return paramdata

# ================================================================
def standardizeWidth( string ):

    global columnWidthMin
    
    while len(string) < columnWidthMin:
        string = ' ' + string
    return string

# ================================================================
def standardizeWidthOfName( string ):

    global columnWidthName
    
    while len(string) < columnWidthName:
        string += ' '
    return string

# ================================================================
def removeParentheses( line ):

    if line.startswith('(') and line.endswith(')'):
        line = line[1:-1]
    return line

# ================================================================
def partInsideOrOutsideParentheses( line, insideOrOutside ):

    idxOpen = line.find( '(' )
    if idxOpen < 0:
        partInside = ''
        partOutside = line
    else:
        idxClose = line.find( ')' )
        if idxClose < 0:
            idxClose = len(line)
        partInside  = line[idxOpen+1:idxClose]
        partsOutside = []
        if idxOpen > 0: 
            partsOutside.append( line[:idxOpen] )
        if idxClose+1 < len(line):
            partsOutside.append( line[idxClose+1:] )
        partOutside = ''.join( partsOutside )

    if insideOrOutside == 'inside':
        return standardizeWidth( partInside.strip() )
    else:
        return standardizeWidth( partOutside.strip() )

# ================================================================
#   some entries in the .csv file are much wider than others,
#   which is inconvenient when viewing the .csv file in a plain text editor.
#   This function breaks a few of those long-entry cases into two rows.
#   This is done through post-processing of the output strings
#   because it violates the one-parameter-per-row rule.
#
def writeAsTwoRows( fid, paramID, row ):

    success = False
    if paramID in ['245-5', '410-3']: # sample (or piano model) ID/name
        success = True
        words = row[0].split('ID/name')
        row1 = [standardizeWidthOfName( words[0] + 'ID'   )]
        row2 = [standardizeWidthOfName( words[0] + 'name' )]
        for n in range(1,len(row)):
            row1.append( partInsideOrOutsideParentheses( row[n], 'outside' ) )
            row2.append( partInsideOrOutsideParentheses( row[n], 'inside' ) )
    elif paramID in ['408-7']:  # sample slot
        success = True
        row1 = [standardizeWidthOfName( row[0] )]
        row2 = [standardizeWidthOfName( 'sample category name' )]
        for n in range(1,len(row)):
            if row[n].strip() == '':
                row1.append( '' )
                row2.append( '' )
            else:
                words = row[n].split( 'in cat' )
                row1.append( standardizeWidth( words[0].strip() ) )
                row2.append( standardizeWidth( words[1].strip() ) )

    if success:
        fid.write( ','.join(row1) )
        fid.write( '\n' )
        fid.write( ','.join(row2) )
        fid.write( '\n' )
    else:
        fid.write( ','.join(row) )
        fid.write( '\n' )

# ================================================================
def writeToolVersion( fid ):

    entries = [standardizeWidthOfName(ns4help.execname + ' version'), \
               standardizeWidth(str(ns4maps.version))]
    fid.write( ','.join(entries) )
    fid.write( '\n' )

# ================================================================
def writeFile( filename, paramdata, groupparams, group, writecsv ):

    global columnWidthMax
    
    testing = False

    paramdata = populateDependson( paramdata )
    numlayers = ns4maps.getNumLayers( group )

    if writecsv:
        headers = [standardizeWidthOfName('parameter')]
        if numlayers > 1:
            for k in range(numlayers):
                headers.append( standardizeWidth( 'layer ' + ns4names.Letter[k] ) )
        else:
            headers.append( standardizeWidth( 'value' ) )
    
        if verbosity >= 2:
           print( 'Writing', filename )
    
        fid = open( filename, 'w' )
        fid.write( ','.join(headers) )
        fid.write( '\n' )
        if group == 'm':
            writeToolVersion( fid )
    for n in range(len(groupparams[group])):

        paramID = groupparams[group][n]

        if not shouldKeepThisParam( paramID ):
            continue

        row = [standardizeWidthOfName( paramdata[paramID]['name'] )]

        dependsonIDs = paramdata[paramID]['dependson']

        if testing:
            print( 'Doing', paramdata[paramID]['name'] )

        columnIsTooWide = False
        for k in range(paramdata[paramID]['numlayers']):

            #  get values of parameters on which this parameter's interpretation 
            #  depends, if any
            dependvals = {}
            for j in range(len(dependsonIDs)):

                #  determine which layer holds the other parameter
                kother = k
                if paramdata[dependsonIDs[j]]['numlayers'] < 2:
                    kother = 0

                dependvals[dependsonIDs[j]] = paramdata[dependsonIDs[j]][kother]['rawval']
            if testing:
                print( 'dependvals =', dependvals )

            interpretedParam = interpretParam(  paramID, \
                                                paramdata[paramID][k]['rawval'], \
                                                dependvals )
            paramdata[paramID][k]['outval'] = interpretedParam # for gui
            if not paramdata[paramID][k]['show']:
                interpretedParam = ''
            row.append( standardizeWidth( interpretedParam ) )

            if len(interpretedParam) > columnWidthMax:
                columnIsTooWide = True

        if writecsv:
            if columnIsTooWide:
                writeAsTwoRows( fid, paramID, row )
            else:
                fid.write( ','.join(row) )
                fid.write( '\n' )
    if writecsv:
        fid.close()

    return paramdata # for gui

# ================================================================
def getLocationCode( paramdata ):

    result = ''
    result += interpretBank( paramdata['013-6'][0]['rawval'] )
    result += ':'
    result += interpretLocnInBank( paramdata['015-3'][0]['rawval'] )
    return result

# ================================================================
def getLayers( numlayers, chosenlayers ):

    if chosenlayers == '':
        layers = range(numlayers)
    else:
        words = chosenlayers.split(',')
        layers = []
        for word in words:
            if word.strip() == 'A':
                layers.append( 0 )
            elif word.strip() == 'B':
                if numlayers > 0:
                    layers.append( 1 )
            elif word.strip() == 'C':
                if numlayers > 1:
                    layers.append( 2 )
    return layers

# ================================================================
def showheader( paramdata, groupparams, options ):

    line = [standardizeWidth( 'program' )]
    for group in options['chosenparams'].keys():
        numlayers = ns4maps.getNumLayers( group )
        groupname = ''
        if group == 'o':
            groupname = 'organ'
        elif group == 'p':
            groupname = 'piano'
        elif group == 'y':
            groupname = 'synth'

        for n in range(len(options['chosenparams'][group])):

            layers = getLayers( numlayers, options['chosenlayers'] )

            paramindex = options['chosenparams'][group][n]
            if paramindex >= len(groupparams[group]):
                print( 'Error: index', paramindex, 'is out of bounds for option -'+group )
                exit()
            paramID = groupparams[group][paramindex]

            for k in layers:
                entry = paramdata[paramID]['name']
                if numlayers > 1:
                    entry = groupname + ' ' + ns4names.Letter[k] + ' ' + entry
                line.append( standardizeWidth( entry ) )
    line.append( 'filename' )
    line = ', '.join(line)

    if 'chosenparamsFileId' in options.keys():
        options['chosenparamsFileId'].write( line + '\n' )
    else:
        print( line )

# ================================================================
def removeExtension( filename ):

    words = filename.split('.')
    return '.'.join(words[:-1])

# ================================================================
def showline( paramdata, groupparams, filename, options ):

    line = standardizeWidth( getLocationCode( paramdata ) )
    for group in options['chosenparams'].keys():
        numlayers = ns4maps.getNumLayers( group )
        for n in range(len(options['chosenparams'][group])):

            paramindex = options['chosenparams'][group][n]

            paramdata = populateDependson( paramdata )
            paramID = groupparams[group][paramindex]
            if not shouldKeepThisParam( paramID ):
                print( 'Error: you requested -' + paramdata[paramID]['group'], paramdata[paramID]['name'] + ', but that is not included in "' + filename + '"' )
                exit()
            dependsonIDs = paramdata[paramID]['dependson']

            layers = getLayers( numlayers, options['chosenlayers'] )

            for k in layers:

                #  get values of parameters on which this parameter's interpretation 
                #  depends, if any
                dependvals = {}
                for j in range(len(dependsonIDs)):
                    kother = k
                    if paramdata[dependsonIDs[j]]['numlayers'] < 2:
                        kother = 0
                    dependvals[dependsonIDs[j]] = paramdata[dependsonIDs[j]][kother]['rawval']
    
                interpretedParam = ''
                if paramdata[paramID][k]['show']:
                    interpretedParam = interpretParam( \
                                             paramID, \
                                             paramdata[paramID][k]['rawval'], \
                                             dependvals )

                line += ', ' + standardizeWidth( interpretedParam )
    line += ', ' + removeExtension( filename )

    if 'chosenparamsFileId' in options.keys():
        options['chosenparamsFileId'].write( line + '\n' )
    else:
        print( line )

# ================================================================
def showListOfParams( paramdata, grouppparams, withcode ):

    for group in grouppparams.keys():
        if group == 'm':
            print( '---------------------------------------' )
            print( 'Parameters not tied to specific layers:' )
            print( '---------------------------------------' )
        elif group == 'o':
            print( '--------------------------------' )
            print( 'Parameters for each organ layer:' )
            print( '--------------------------------' )
        elif group == 'p':
            print( '--------------------------------' )
            print( 'Parameters for each piano layer:' )
            print( '--------------------------------' )
        elif group == 'y':
            print( '--------------------------------' )
            print( 'Parameters for each synth layer:' )
            print( '--------------------------------' )
        for n in range(len(grouppparams[group])):
            paramID = grouppparams[group][n]
            if withcode:
                print( '-' + group, f"{n:3d}", paramdata[paramID]['name'] )
            else:
                # this version is for posting the parameter-list on the website
                print( paramdata[paramID]['name'] )

# ================================================================
def setFileVersion( paramdata ):

    global fileVersion
    global warnAboutOlderFileVersion
    global warnAboutNewerFileVersion

    fileVersionID = ns4maps.shouldInterpretAs['fileVersion'][0]
    fileVersion = interpretFileVersion( paramdata[fileVersionID][0]['rawval'] )

    if warnAboutOlderFileVersion:
        if verbosity > 0 and \
           fileVersionOlderThan( ns4maps.fileVersionOldestTested ):
            print( '' )
            print( '*** WARNING ***' )
            print( 'One or more of the ns4 files being decoded is older than any', \
                    'of the versions on which', ns4help.execname, \
                    'has been tested, so the results might not be reliable.', \
                    'The User Manual explains how to fix this.' )
            print( '' )
            warnAboutOlderFileVersion = False  # don't warn again

    if warnAboutNewerFileVersion:
        if verbosity > 0 and \
           fileVersionNewerThan( ns4maps.fileVersionNewestTested ):
            print( '' )
            print( '*** WARNING ***' )
            print( 'One or more of the ns4 files being decoded is newer than any', \
                    'of the versions on which', ns4help.execname, \
                    'has been tested, so the results might not be reliable.' )
            print( '' )
            warnAboutNewerFileVersion = False  # don't warn again

# ================================================================
def doOneFile( paramdata, groupparams, inputFilename, outputFilenameSet, options ):

    global relevantGroups
    global fileType

    if verbosity >= 1:
        print( 'Decoding', inputFilename )

    paramdata = readFile( inputFilename, paramdata )
    setFileVersion( paramdata )

    paramdata = ns4exclude.initialize( paramdata )
    if not options['includeall']:
        paramdata = ns4exclude.decideWhetherToExclude( paramdata, dontKeep )

    # https://stackoverflow.com/questions/23177439/how-to-check-if-a-dictionary-is-empty
    if not outputFilenameSet:
        #  this is the case where the user specified a list of parameters to view,
        #  so the output has one row per file
        showline( paramdata, groupparams, inputFilename, options )
    else:
        #  this is the default case: a set of output files is written for each input file
        for group in ns4maps.groups:
            if group in relevantGroups[fileType]:
                paramdata = writeFile( outputFilenameSet[group], paramdata, \
                                       groupparams, group, options['csv'] )
        if options['gui']:
            ns4gui.showGUI( inputFilename, paramdata, groupparams )

# ================================================================
def constructOutputFilenameSets( options ):

    outputFilenameSets = []
    for n in range(len(options['inputFilenames'])):

        outputFilenameSet = {}
        if len(options['chosenparams'].keys()) == 0:

            words = options['inputFilenames'][n].split('.')
            prefix = '.'.join( words[:-1] ) # input filename without extension
            if options['outputDirectory'] != '':
                prefix = ospath.join( options['outputDirectory'], ospath.basename( prefix ) )
            outputFilenameSet['m'] = ''.join( [prefix, '_master.csv'] )
            outputFilenameSet['o'] = ''.join( [prefix, '_organ.csv'] )
            outputFilenameSet['p'] = ''.join( [prefix, '_piano.csv'] )
            outputFilenameSet['y'] = ''.join( [prefix, '_synth.csv'] )

        outputFilenameSets.append( outputFilenameSet )

    return outputFilenameSets

# ================================================================
def warnAboutOverwriting( outputFilenameSets ):

    for n in range(len(outputFilenameSets)):
        for group in outputFilenameSets[n].keys():
            filename = outputFilenameSets[n][group]
            if ospath.isfile( filename ):
                print( 'Warning: one or more of the .csv files to be written has the same name as an existing file, which will be overwritten. (Example:', filename, ')' )
                yesorno = input( 'Do you want to proceed (y/n)? ' )
                if yesorno.strip() in ['y', 'yes', 'Y', 'Yes', 'YES']:
                    return
                else:
                    print( 'Exiting without writing any output files.' )
                    exit()

# ================================================================
#   main program
#

options = ns4options.getOptions()

verbosity                   = options['verbosity']
warnAboutOlderFileVersion   = options['warnversion']
warnAboutNewerFileVersion   = options['warnversion']
valueformat                 = options['valueformat']
columnWidthMin              = options['columnwidth']

paramdata, groupparams = ns4maps.getMaps(     options['formatspath'], verbosity )
pianomodels, samplenames = ns4names.getNames( options['formatspath'], verbosity )
populate_presetPart_info()

if options['listparams']:
    showListOfParams( paramdata, groupparams, True )
    exit()
if options['listparamsnocodes']:
    showListOfParams( paramdata, groupparams, False )
    exit()
if options['writenames']:
    ns4names.writeNames( options['outputDirectory'], pianomodels, samplenames, verbosity )
if options['writemaps']:
    ns4maps.writeMaps( options['outputDirectory'], paramdata, groupparams, verbosity )
if options['writenames'] or options['writemaps']:
    if len(options['inputFilenames']) > 0:
        print( 'Warning: ns4 files will not be processed when the --names or --bitmaps option is used' )
    #  this prevents an awkward situation where different kinds of outputs
    #  are written to different places.
    #  I didn't want to have to explain that in the where-things-were-written announcement.
    exit()

outputFilenameSets = constructOutputFilenameSets( options )

if options['gui'] and len(options['inputFilenames']) > 0 and not outputFilenameSets[0]:
    print( 'Error: cannot use option --gui when requesting specific parameters' )
    exit()

if options['askfirst'] and options['csv']:
    warnAboutOverwriting( outputFilenameSets )

if len(options['chosenparams'].keys()) > 0:
    #  if the user specified a list of parameters to view,
    #  then the output will consist of one row per file (instead of one per parameter),
    #  so the .csv header needs to be written before the loop over files
    if options['chosenparamsFile'] != "":
        # announce output regardless of verbosity,
        # because in this mode the verbosity is automatically set to 0 in ns4options.py
        print( 'Writing', options['chosenparamsFile'] )
        options['chosenparamsFileId'] = open( options['chosenparamsFile'], 'w' )
    showheader( paramdata, groupparams, options )

for n in range(len(options['inputFilenames'])):
    doOneFile( paramdata, groupparams, options['inputFilenames'][n], outputFilenameSets[n], options )

#  tell the user where the output files were written
if options['csv'] and len(options['chosenparams'].keys()) == 0 \
                  and len(options['inputFilenames']) > 0:
    outputDirectory = options['outputDirectory']
    if outputDirectory == '.':
        outputDirectory = 'current working directory'
    elif outputDirectory == '':
        outputDirectory = 'the directory(s) that host the input files'
    print( 'Outputs written to', outputDirectory )

if 'chosenparamsFileId' in options.keys():
    options['chosenparamsFileId'].close()
