#
#  -------------------------------------------------------------------------
#  MIT License
#  
#  Copyright (c) 2024 Randy
#  
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  -------------------------------------------------------------------------
#
#  This module provides the function decideWhetherToExclude(),
#  which sets the 'show' field of some parameters to False
#  based on the on/off status of the section/layer/collection in which it lives
#  on the Nord Stage 4 front panel
#
import ns4maps

# ================================================================
def getCollection_organSection( paramdata ):

    # WARNING: this is fragile because it depends on the name
    prefix = 'organ FX'

    collection = []
    for paramID in paramdata.keys():
        if paramdata[paramID]['group'] == 'o':
            collection.append( paramID )
        elif paramdata[paramID]['group'] == 'm' \
         and paramdata[paramID]['name'].startswith( prefix ):
            collection.append( paramID )

    return collection

# ================================================================
def getCollection_pianoSection( paramdata ):

    collection = []
    for paramID in paramdata.keys():
        if paramdata[paramID]['group'] == 'p':
            collection.append( paramID )

    return collection

# ================================================================
def getCollection_synthSection( paramdata ):

    collection = []
    for paramID in paramdata.keys():
        if paramdata[paramID]['group'] == 'y':
            collection.append( paramID )

    return collection

# ================================================================
def getCollection_layer( paramdata, group, switchIDs ):

    collection = []
    for paramID in paramdata.keys():
        if paramdata[paramID]['group'] == group:
            if not (paramID in switchIDs): # still show the on/off switch
                collection.append( paramID )

    return collection

# ================================================================
def getCollection( paramdata, group, switchID, prefix ):

    collection = []
    for paramID in paramdata.keys():
        if paramdata[paramID]['group'] == group and \
           paramdata[paramID]['name'].startswith( prefix ):
            if not (paramID == switchID): # still show the on/off switch
                collection.append( paramID )

    return collection

# ================================================================
def enforceSwitch( paramdata, group, switchID, prefix, rawvaloff ):

    collection = getCollection( paramdata, group, switchID, prefix )

    for paramID in collection:
        numlayers = ns4maps.getNumLayers( group )
        for k in range(numlayers):
            if paramdata[switchID][k]['rawval'] == rawvaloff:
                paramdata[paramID][k]['show'] = False

    return paramdata

# ================================================================
def enforceSwitchFX( paramdata, group, prefix ):

    if paramdata['088-4'][0]['rawval'] == 0: # toggles FX for all sections/layers

        collection = getCollection( paramdata, group, '', prefix )
    
        for paramID in collection:
            numlayers = ns4maps.getNumLayers( group )
            for k in range(numlayers):
                paramdata[paramID][k]['show'] = False

    return paramdata

# ================================================================
def doFX( paramdata, group, dontKeep ):

     # WARNING: prefix is fragile because it depends on the name

    allFx = '088-4'

    preprefix = ''
    realgroup = group
    if group == 'm':
        # if the group is 'm', then assume we're doing organ FX parameters
        preprefix = 'organ '
        realgroup = 'o'

    prefix = preprefix + 'FX mod 1 '
    switchID = ns4maps.shiftBit( '705-1', ns4maps.bitsOffsetFX[realgroup] )
    if isAvailable( [switchID,allFx], dontKeep ):
        paramdata = enforceSwitch( paramdata, group, switchID, prefix, 0 )
        paramdata = enforceSwitchFX( paramdata, group, prefix )

    prefix = preprefix + 'FX mod 2 '
    switchID = ns4maps.shiftBit( '713-5', ns4maps.bitsOffsetFX[realgroup] )
    if isAvailable( [switchID,allFx], dontKeep ):
        paramdata = enforceSwitch( paramdata, group, switchID, prefix, 0 )
        paramdata = enforceSwitchFX( paramdata, group, prefix )

    prefix = preprefix + 'FX amp sim/EQ '
    switchID = ns4maps.shiftBit( '721-8', ns4maps.bitsOffsetFX[realgroup] )
    if isAvailable( [switchID,allFx], dontKeep ):
        paramdata = enforceSwitch( paramdata, group, switchID, prefix, 0 )
        paramdata = enforceSwitchFX( paramdata, group, prefix )

    prefix = preprefix + 'FX comp '
    switchID = ns4maps.shiftBit( '732-8', ns4maps.bitsOffsetFX[realgroup] )
    if isAvailable( [switchID,allFx], dontKeep ):
        paramdata = enforceSwitch( paramdata, group, switchID, prefix, 0 )
        paramdata = enforceSwitchFX( paramdata, group, prefix )

    prefix = preprefix + 'FX delay '
    switchID = ns4maps.shiftBit( '734-1', ns4maps.bitsOffsetFX[realgroup] )
    if isAvailable( [switchID,allFx], dontKeep ):
        paramdata = enforceSwitch( paramdata, group, switchID, prefix, 0 )
        paramdata = enforceSwitchFX( paramdata, group, prefix )

    prefix = preprefix + 'FX reverb '
    switchID = ns4maps.shiftBit( '750-4', ns4maps.bitsOffsetFX[realgroup] )
    if isAvailable( [switchID,allFx], dontKeep ):
        paramdata = enforceSwitch( paramdata, group, switchID, prefix, 0 )
        paramdata = enforceSwitchFX( paramdata, group, prefix )

    return paramdata

# ================================================================
def initialize( paramdata ):

    for paramID in paramdata.keys():
        numlayers = ns4maps.getNumLayers( paramdata[paramID]['group'] )
        for k in range(numlayers):
            paramdata[paramID][k]['show'] = True

    return paramdata

# ================================================================
def isAvailable( paramIDs, dontKeep ):

    for paramID in paramIDs:
        if paramID in dontKeep:
            return False
    return True

# ================================================================
#   assumes that each parameter's 'rawval' field is already populated
#   and that the 'show' field is initially set to True
#
def decideWhetherToExclude( paramdata, dontKeep ):

    if isAvailable( ['084-5', '084-8'], dontKeep ): # not in ns4n,ns4y
        if (paramdata['084-5'][0]['rawval'] == 0) and \
           (paramdata['084-8'][0]['rawval'] == 0):
            # section is off in both scenes
            collection = getCollection_organSection( paramdata )
            for paramID in collection:
                numlayers = ns4maps.getNumLayers( paramdata[paramID]['group'] )
                for k in range(numlayers):
                    paramdata[paramID][k]['show'] = False
            # a few organ parameters are in the 'm' group
            paramdata['103-5'][0]['show'] = False
            paramdata['104-1'][0]['show'] = False
            paramdata['110-3'][0]['show'] = False

    if isAvailable( ['084-6', '085-1'], dontKeep ): # not in ns4o,ns4y
        if (paramdata['084-6'][0]['rawval'] == 0) and \
           (paramdata['085-1'][0]['rawval'] == 0):
            # section is off in both scenes
            collection = getCollection_pianoSection( paramdata )
            for paramID in collection:
                numlayers = ns4maps.getNumLayers( paramdata[paramID]['group'] )
                for k in range(numlayers):
                    paramdata[paramID][k]['show'] = False

    if isAvailable( ['084-7', '085-2'], dontKeep ): # not in ns4o,ns4n
        if (paramdata['084-7'][0]['rawval'] == 0) and \
           (paramdata['085-2'][0]['rawval'] == 0):
            # section is off in both scenes
            collection = getCollection_synthSection( paramdata )
            for paramID in collection:
                numlayers = ns4maps.getNumLayers( paramdata[paramID]['group'] )
                for k in range(numlayers):
                    paramdata[paramID][k]['show'] = False

    # organ layers
    switchIDs = ['095-3', '095-6']
    if isAvailable( switchIDs, dontKeep ):
        collection = getCollection_layer( paramdata, 'o', switchIDs )
        for paramID in collection:
            numlayers = ns4maps.getNumLayers( paramdata[paramID]['group'] )
            for k in range(numlayers):
                if (paramdata['095-3'][k]['rawval'] == 0) and \
                   (paramdata['095-6'][k]['rawval'] == 0):
                    # layer is off in both scenes
                    paramdata[paramID][k]['show'] = False

    # piano layers
    switchIDs = ['230-3', '230-6']
    if isAvailable( switchIDs, dontKeep ):
        collection = getCollection_layer( paramdata, 'p', switchIDs )
        for paramID in collection:
            numlayers = ns4maps.getNumLayers( paramdata[paramID]['group'] )
            for k in range(numlayers):
                if (paramdata['230-3'][k]['rawval'] == 0) and \
                   (paramdata['230-6'][k]['rawval'] == 0):
                    # layer is off in both scenes
                    paramdata[paramID][k]['show'] = False

    # synth layers
    switchIDs = ['377-5', '378-2']
    if isAvailable( switchIDs, dontKeep ):
        collection = getCollection_layer( paramdata, 'y', switchIDs )
        for paramID in collection:
            numlayers = ns4maps.getNumLayers( paramdata[paramID]['group'] )
            for k in range(numlayers):
                if (paramdata['377-5'][k]['rawval'] == 0) and \
                   (paramdata['378-2'][k]['rawval'] == 0):
                    # layer is off in both scenes
                    paramdata[paramID][k]['show'] = False

    prefix = 'KB zone' # WARNING: prefix is fragile because it depends on the name
    switchID = '050-1'
    if isAvailable( [switchID], dontKeep ):
        for group in ['m', 'o', 'p', 'y']:
            collection = getCollection( paramdata, group, switchID, prefix )
        
            for paramID in collection:
                numlayers = ns4maps.getNumLayers( group )
                for k in range(numlayers):
                    if paramdata[switchID][0]['rawval'] == 0:
                        paramdata[paramID][k]['show'] = False

    prefix = 'program transpose ' # WARNING: prefix is fragile because it depends on the name
    switchID = '052-7'
    if isAvailable( [switchID], dontKeep ):
        group = 'm'
        paramdata = enforceSwitch( paramdata, group, switchID, prefix, 0 )

    # WARNING: prefix is fragile because it depends on the name

    prefix = 'sample ' 
    switchID = '408-1'
    if isAvailable( [switchID], dontKeep ):
        group = 'y'
        paramdata = enforceSwitch( paramdata, group, switchID, prefix, 1 ) # using analog

    prefix = 'analog ' 
    switchID = '408-1'
    if isAvailable( [switchID], dontKeep ):
        group = 'y'
        paramdata = enforceSwitch( paramdata, group, switchID, prefix, 0 ) # using sample

#    prefix = 'osc ctrl' 
#    switchID = '563-2'
#    if isAvailable( [switchID], dontKeep ):
#        group = 'y'
#        paramdata = enforceSwitch( paramdata, group, switchID, prefix, 0 ) # using sample

#    prefix = 'osc env' 
#    switchID = '571-5'
#    if isAvailable( [switchID], dontKeep ):
#        group = 'y'
#        paramdata = enforceSwitch( paramdata, group, switchID, prefix, 0 ) # using sample

    prefix = 'extern ' 
    switchID = '418-1'
    if isAvailable( [switchID], dontKeep ):
        group = 'y'
        paramdata = enforceSwitch( paramdata, group, switchID, prefix, 0 )

    prefix = 'arpeggiator ' 
    switchID = '423-8'
    if isAvailable( [switchID], dontKeep ):
        group = 'y'
        paramdata = enforceSwitch( paramdata, group, switchID, prefix, 0 )

    prefix = 'arp ' 
    switchID = '423-8'
    if isAvailable( [switchID], dontKeep ):
        group = 'y'
        paramdata = enforceSwitch( paramdata, group, switchID, prefix, 0 )

    if isAvailable( ['424-3'], dontKeep ):
        for layer in [0,1,2]: # arpeggiator pattern on/off
            if paramdata['424-3'][layer]['rawval'] == 0:
                paramdata['432-7'][layer]['show'] = False
                paramdata['433-3'][layer]['show'] = False
                paramdata['437-3'][layer]['show'] = False
                paramdata['441-3'][layer]['show'] = False

    prefix = 'LFO ' 
    switchID = '575-8'
    if isAvailable( [switchID], dontKeep ):
        group = 'y'
        paramdata = enforceSwitch( paramdata, group, switchID, prefix, 0 )

    prefix = 'filter ' 
    switchID = '602-5'
    if isAvailable( [switchID], dontKeep ):
        group = 'y'
        paramdata = enforceSwitch( paramdata, group, switchID, prefix, 0 )

    for group in ['m', 'p', 'y']:
        paramdata = doFX( paramdata, group, dontKeep )

    return paramdata
