# ================================================================
#
#  -------------------------------------------------------------------------
#  MIT License
#  
#  Copyright (c) 2024 Randy
#  
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  -------------------------------------------------------------------------
#
#  this module is imported by ns4decode.py
#
#  this module provides the getMaps() function, which returns information
#  about where each parameter lives in an Program (ns4p) file
#
#  the adjustments needed for Preset files (ns4o, ns4n, ns4y) are defined in ns4decode.py
#
#  the main data structures are paramdata and groupparams.
#
#  paramdata is a dictionary whose key is the ID number of a parameter.
#  The ID numaber of a parameter is a string of the form XXX-Y
#  that specifies the location of first bit of the first layer's version of that parameter.
#
#  paramdata[id] is a dictionary with these fields:
#
#       paramdata[id][layer index]['begloc'] = index of bit-location
#       paramdata[id][layer index]['endloc'] = index of bit-location
#       paramdata[id][layer index]['rawval'] = raw (not interpreted) integer value
#       paramdata[id][layer index]['outval'] = interpreted value (a string)
#       paramdata[id]['name'] = string
#       paramdata[id]['dependson'] = list of ids (the significance of the ordering 
#                                                 depends on the parameter)
#       paramdata[id]['group'] = 'm', 'o', 'p', or 's'
#           (for master, organ, piano, and synth)
#       paramdata[id]['numlayers'] = number of layers in this group
#
#  groupparams is a dictionary whose keys are 'm', 'o', 'p', and 's'.
#  Each of those is a list of id-strings.
#  This is used to specify the ordering of the parameters in the output.
#
#  The parameters listed in paramdata[id]['dependson'] must
#  belong either to groupparams[paramdata[id]['group']]
#  or to groupparams['m'].
# ================================================================
from os import path as ospath
from sys import exit

import ns4names

mapFilenames = {}
mapFilenames['m'] = 'ns4p_bitmap_master.txt'
mapFilenames['o'] = 'ns4p_bitmap_organ.txt'
mapFilenames['p'] = 'ns4p_bitmap_piano.txt'
mapFilenames['y'] = 'ns4p_bitmap_synth.txt'

groups = list( mapFilenames.keys() )

version = '2.0' # version of ns4decode

#  fileVersionChanges[] tries to keep track of 
#  when various changes were introduced into the
#  ns4p file format (controlled by Nord, not by ns4decode)
#
#  Based on a single test-case from 2024-09-24, 
#  version 3.08 seems to be the same as 3.07, and
#  version 3.09 seems to be the same as 3.13, at least for SynthTypes
fileVersionChanged = {}

#  sometime after 3.07, but not sure when (probably OS version 1.32, 
#  but don't know the corresponding file version)
fileVersionChanged['OrganVibChorusType'] = {}
fileVersionChanged['OrganVibChorusType']['ns4p'] = '3.13' 
fileVersionChanged['OrganVibChorusType']['ns4o'] = '2.01' # maybe earlier

fileVersionChanged['SynthTypes'] = {}
fileVersionChanged['SynthTypes']['ns4p'] = '3.09' # introduced in NS4 OS version 1.20
fileVersionChanged['SynthTypes']['ns4y'] = '2.04' # after 2.03 but not after 2.08

# sometime after 3.07, but not sure when 
# (probably OS version 1.44, based on https://www.norduserforum.com/viewtopic.php?t=25546)
fileVersionChanged['VibratoDelay'] = {}
fileVersionChanged['VibratoDelay']['ns4p'] = '3.13'
fileVersionChanged['VibratoDelay']['ns4y'] = '2.03' # maybe earlier

fileVersionOldestTested = {} # first version on which ns4decode was tested at all
fileVersionNewestTested = {} # first version on which ns4decode was extensively tested

fileVersionOldestTested['ns4p'] = '3.04'
fileVersionNewestTested['ns4p'] = '3.13'

fileVersionOldestTested['ns4o'] = '2.01'
fileVersionNewestTested['ns4o'] = '2.01'

fileVersionOldestTested['ns4n'] = '2.01'
fileVersionNewestTested['ns4n'] = '2.01'

fileVersionOldestTested['ns4y'] = '2.03'
fileVersionNewestTested['ns4y'] = '2.03'

bitsOffsetFX = {}
bitsOffsetFX['o'] = -530*8
bitsOffsetFX['p'] = -438*8
bitsOffsetFX['y'] = 0

# ================================================================
#   essentially the inverse of getLocs(), but for just beg or end (not both)
#
def makeLocStr( loc ):

    locbyte = 1 + int(loc/8)
    locbit  = 1 + loc - (locbyte-1)*8

    return f"{locbyte:03d}-{locbit:1d}"

# ================================================================
def shiftBit( oldID, shiftAmount ):

    return makeLocStr( getLocFromStr( oldID ) + shiftAmount )

# ================================================================
#  byteloc = the location (1-based) of the byte in which the bit occurs
#  bitlocinbyte = the location (1-based) of the bit within the byte
#  result = the location (0-based) of the bit in a contiguous string of bits
#
def getBitloc( byteloc, bitlocinbyte ):

    return (byteloc-1)*8 + (bitlocinbyte-1)

# ================================================================
#  the format of locstr is XXX-Y,
#  where XXX is a 3-digit 1-based decimal number specifying the byte,
#  and Y is a 1-digit 1-based decimal number specifying the bit within that byte
#
def getLocFromStr( locstr ):

    return getBitloc( int(locstr[:3]), int(locstr[4:]) )

# ================================================================
#  input is two words, representing the most- and least-significant bits
#
def getLocs( beglocstr, endlocstr, offset ):

    result = {}
    if beglocstr != '':
        if endlocstr.strip() == '':
            endlocstr = beglocstr
        result['begloc'] = getLocFromStr( beglocstr ) + offset
        result['endloc'] = getLocFromStr( endlocstr ) + offset
    else:
        result['begloc'] = -1
        result['endloc'] = -1

    return result

# ================================================================
def getNumLayers( group ):

    numlayers = 1
    if group == 'o' or group == 'p':  # organ or piano 
        numlayers = 2
    elif group == 'y':  # synth
        numlayers = 3

    return numlayers

# ================================================================
#   given information about where to find a parameter in the ns4p file format,
#   this function adds that information to the paramdata and groupparams structures
#
def addParam( group, groupoffset, layeroffset, beglocstr, endlocstr, name ):

    global paramdata
    global groupparams
    
    if group == 'o' and name.startswith( 'FX' ):
        # prepend 'organ' to the name because both organ layers share the same FX,
        # so the organ FX params are associated with the 'm' (main) group
        # because it's the only group that doesn't have multiple layers
        name = 'organ ' + name
        group = 'm'

    #  using groupoffset here ensures that FX params 
    #  in different sections have different IDs
    paramID = shiftBit( beglocstr, groupoffset )

    numlayers = getNumLayers( group )

    paramdata[paramID] = {}
    paramdata[paramID]['name'] = name
    paramdata[paramID]['group'] = group
    paramdata[paramID]['numlayers'] = numlayers
    paramdata[paramID]['dependson'] = [] # will be populated later

    for k in range(numlayers):
        locs = getLocs( beglocstr, endlocstr, k * layeroffset + groupoffset )
        paramdata[paramID][k] = {}
        paramdata[paramID][k]['begloc'] = locs['begloc']
        paramdata[paramID][k]['endloc'] = locs['endloc']

    groupparams[group].append( paramID )

# ================================================================
#  the group offset (go) should be relative to layer 1 of the synth section,
#  which is what the beglocstr and endlocstr in this list refer to
#
def addParamsFX( grp, go ):

    lo = 55*8
    addParam( grp, go, lo, '705-1', '     ', 'FX mod 1 on/off' )
    addParam( grp, go, lo, '705-2', '     ', 'FX mod 1 MST CLK on/off' )
    addParam( grp, go, lo, '705-3', '706-1', 'FX mod 1 rate' )
    addParam( grp, go, lo, '706-2', '707-1', 'FX mod 1 rate with wheel' )
    addParam( grp, go, lo, '707-2', '708-1', 'FX mod 1 rate with A.T.' )
    addParam( grp, go, lo, '708-2', '709-1', 'FX mod 1 rate with ctrlped' )
    addParam( grp, go, lo, '709-2', '709-8', 'FX mod 1 amount' )
    addParam( grp, go, lo, '710-1', '710-8', 'FX mod 1 amount with wheel' )
    addParam( grp, go, lo, '711-1', '711-8', 'FX mod 1 amount with A.T.' )
    addParam( grp, go, lo, '712-1', '712-8', 'FX mod 1 amount with ctrlped' )
    addParam( grp, go, lo, '713-1', '713-4', 'FX mod 1 mode' )

    addParam( grp, go, lo, '713-5', '     ', 'FX mod 2 on/off' )
    addParam( grp, go, lo, '713-6', '714-4', 'FX mod 2 rate' )
    addParam( grp, go, lo, '714-5', '715-4', 'FX mod 2 rate with wheel' )
    addParam( grp, go, lo, '715-5', '716-4', 'FX mod 2 rate with A.T.' )
    addParam( grp, go, lo, '716-5', '717-4', 'FX mod 2 rate with ctrlped' )
    addParam( grp, go, lo, '717-5', '718-3', 'FX mod 2 amount' )
    addParam( grp, go, lo, '718-4', '719-3', 'FX mod 2 amount with wheel' )
    addParam( grp, go, lo, '719-4', '720-3', 'FX mod 2 amount with A.T.' )
    addParam( grp, go, lo, '720-4', '721-3', 'FX mod 2 amount with ctrlped' )
    addParam( grp, go, lo, '721-4', '721-7', 'FX mod 2 mode' )

    addParam( grp, go, lo, '721-8', '     ', 'FX amp sim/EQ on/off' )
    addParam( grp, go, lo, '722-1', '722-7', 'FX amp sim/EQ treb' )
    addParam( grp, go, lo, '722-8', '723-6', 'FX amp sim/EQ mid' )
    addParam( grp, go, lo, '723-7', '724-5', 'FX amp sim/EQ bass' )
    addParam( grp, go, lo, '724-6', '725-4', 'FX amp sim/EQ freq' )
    addParam( grp, go, lo, '725-5', '726-4', 'FX amp sim/EQ freq with wheel' )
    addParam( grp, go, lo, '726-5', '727-4', 'FX amp sim/EQ freq with A.T.' )
    addParam( grp, go, lo, '727-5', '728-4', 'FX amp sim/EQ freq with ctrlped' )
    addParam( grp, go, lo, '728-5', '729-3', 'FX amp sim/EQ drive' )
    addParam( grp, go, lo, '729-4', '730-3', 'FX amp sim/EQ drive with wheel' )
    addParam( grp, go, lo, '730-4', '731-3', 'FX amp sim/EQ drive with A.T.' )
    addParam( grp, go, lo, '731-4', '732-3', 'FX amp sim/EQ drive with ctrlped' )
    addParam( grp, go, lo, '755-6', '756-1', 'FX amp sim/EQ mode' )

    addParam( grp, go, lo, '732-8', '     ', 'FX comp on/off' )
    addParam( grp, go, lo, '733-1', '733-7', 'FX comp amount' )
    addParam( grp, go, lo, '733-8', '     ', 'FX comp response' )

    addParam( grp, go, lo, '734-1', '     ', 'FX delay on/off' )
    addParam( grp, go, lo, '734-2', '     ', 'FX delay tempo MST CLK on/off' )
    addParam( grp, go, lo, '734-3', '735-1', 'FX delay tempo' )
    addParam( grp, go, lo, '736-1', '736-8', 'FX delay tempo with wheel' )
    addParam( grp, go, lo, '737-1', '737-8', 'FX delay tempo with A.T.' )
    addParam( grp, go, lo, '738-1', '738-8', 'FX delay tempo with ctrlped' )
    addParam( grp, go, lo, '741-6', '742-4', 'FX delay mix' )
    addParam( grp, go, lo, '742-5', '743-4', 'FX delay mix with wheel' )
    addParam( grp, go, lo, '743-5', '744-4', 'FX delay mix with A.T.' )
    addParam( grp, go, lo, '744-5', '745-4', 'FX delay mix with ctrlped' )
    addParam( grp, go, lo, '745-5', '     ', 'FX delay normal/analog' )
    addParam( grp, go, lo, '745-6', '     ', 'FX delay ping pong on/off' )
    addParam( grp, go, lo, '745-7', '745-8', 'FX delay filter type' )
    addParam( grp, go, lo, '746-1', '746-7', 'FX delay feedback' )
    addParam( grp, go, lo, '746-8', '747-7', 'FX delay feedback with wheel' )
    addParam( grp, go, lo, '747-8', '748-7', 'FX delay feedback with A.T.' )
    addParam( grp, go, lo, '748-8', '749-7', 'FX delay feedback with ctrlped' )
    addParam( grp, go, lo, '749-8', '750-3', 'FX delay effects' )

    addParam( grp, go, lo, '750-4', '     ', 'FX reverb on/off' )
    addParam( grp, go, lo, '750-5', '751-3', 'FX reverb amount' )
    addParam( grp, go, lo, '751-4', '752-3', 'FX reverb amount with wheel' )
    addParam( grp, go, lo, '752-4', '753-3', 'FX reverb amount with A.T.' )
    addParam( grp, go, lo, '753-4', '754-3', 'FX reverb amount with ctrlped' )
    addParam( grp, go, lo, '754-4', '754-5', 'FX reverb dark/bright' )
    addParam( grp, go, lo, '754-6', '755-1', 'FX reverb type' )

# ================================================================
def makeMapMaster():

    grp = 'm'
    go = 0

    lo = 0

    addParam( grp, go, lo, '009-1', '012-8', 'file type' )
    addParam( grp, go, lo, '021-1', '021-8', 'file version' )
    addParam( grp, go, lo, '048-1', '048-8', 'file version again' )
    addParam( grp, go, lo, '013-6', '013-8', 'bank' )
    addParam( grp, go, lo, '015-3', '015-8', 'location in bank' )
    addParam( grp, go, lo, '025-1', '028-8', 'checksum' )
    addParam( grp, go, lo, '050-1', '     ', 'split on/off' )
    addParam( grp, go, lo, '050-2', '     ', 'KB zones 1-2 split point on/off' )
    addParam( grp, go, lo, '050-3', '     ', 'KB zones 2-3 split point on/off' )
    addParam( grp, go, lo, '050-4', '     ', 'KB zones 3-4 split point on/off' )
    addParam( grp, go, lo, '050-5', '050-8', 'KB zones 1-2 split point' )
    addParam( grp, go, lo, '051-1', '051-4', 'KB zones 2-3 split point' )
    addParam( grp, go, lo, '051-5', '051-8', 'KB zones 3-4 split point' )
    addParam( grp, go, lo, '052-1', '052-2', 'KB zones 1-2 split point Xfade' )
    addParam( grp, go, lo, '052-3', '052-4', 'KB zones 2-3 split point Xfade' )
    addParam( grp, go, lo, '052-5', '052-6', 'KB zones 3-4 split point Xfade' )
    addParam( grp, go, lo, '052-7', '     ', 'program transpose on/off' )
    addParam( grp, go, lo, '052-8', '053-3', 'program transpose amount' )
    addParam( grp, go, lo, '054-8', '     ', 'FX comp global on/off' )
    addParam( grp, go, lo, '055-1', '     ', 'FX delay global on/off' )
    addParam( grp, go, lo, '055-2', '     ', 'FX reverb global on/off' )
#    addParam( grp, go, lo, '056-5', '060-4', 'organ preset ID' )
    addParam( grp, go, lo, '084-5', '     ', 'organ section on/off' )
    addParam( grp, go, lo, '084-6', '     ', 'piano section on/off' )
    addParam( grp, go, lo, '084-7', '     ', 'synth section on/off' )
    addParam( grp, go, lo, '084-8', '     ', 'organ section on/off (layer scene II)' )
    addParam( grp, go, lo, '085-1', '     ', 'piano section on/off (layer scene II)' )
    addParam( grp, go, lo, '085-2', '     ', 'synth section on/off (layer scene II)' )
    addParam( grp, go, lo, '085-3', '     ', 'which layer scene is active' )
    addParam( grp, go, lo, '088-4', '     ', 'FX on/off' )
    addParam( grp, go, lo, '103-5', '     ', 'organ pstick on/off' )
    addParam( grp, go, lo, '104-1', '     ', 'organ rotary spkr on/off' )
    addParam( grp, go, lo, '105-5', '106-3', 'rotary spkr drive' )
    addParam( grp, go, lo, '108-5', '108-7', 'rotary spkr stop position' )
    addParam( grp, go, lo, '109-3', '     ', 'rotary spkr stop on/off' )
    addParam( grp, go, lo, '109-4', '     ', 'rotary spkr slow/fast' )
    addParam( grp, go, lo, '110-3', '111-1', 'organ vib/chorus type' )
    addParam( grp, go, lo, '401-6', '     ', 'synth arp group on/off' )
    addParam( grp, go, lo, '401-8', '     ', 'synth KB hold on/off' )

# ================================================================
def makeMapOrgan():

    grp = 'o'
    go = 0

    lo = -1
    addParam( grp, go, lo, '095-3', '     ', 'layer on/off' )
    addParam( grp, go, lo, '095-6', '     ', 'layer on/off (scene II)' )

    lo = 31
    addParam( grp, go, lo, '095-7', '096-5', 'volume' )
    addParam( grp, go, lo, '096-6', '097-5', 'volume change with wheel' )
    addParam( grp, go, lo, '097-6', '098-5', 'volume change with A.T.' )
    addParam( grp, go, lo, '098-6', '099-5', 'volume change with ctrlped' )

    lo = 31*8
    addParam( grp, go, lo, '113-1', '113-4', 'KB zones' )
    addParam( grp, go, lo, '113-5', '113-8', 'octave shift' )
    addParam( grp, go, lo, '114-1', '     ', 'organ susped on/off' )
    addParam( grp, go, lo, '114-2', '114-4', 'organ model' )
    addParam( grp, go, lo, '114-5', '     ', 'preset on/off' )

    addParam( grp, go, lo, '117-1', '117-4', 'drawbar 1' )
    addParam( grp, go, lo, '117-5', '118-1', 'drawbar 1 with wheel' )
    addParam( grp, go, lo, '118-2', '118-6', 'drawbar 1 with A.T.' )
    addParam( grp, go, lo, '118-7', '119-3', 'drawbar 1 with ctrlped' )

    addParam( grp, go, lo, '119-4', '119-7', 'drawbar 2' )
    addParam( grp, go, lo, '119-8', '120-4', 'drawbar 2 with wheel' )
    addParam( grp, go, lo, '120-5', '121-1', 'drawbar 2 with A.T.' )
    addParam( grp, go, lo, '121-2', '121-6', 'drawbar 2 with ctrlped' )

    addParam( grp, go, lo, '121-7', '122-2', 'drawbar 3' )
    addParam( grp, go, lo, '122-3', '122-7', 'drawbar 3 with wheel' )
    addParam( grp, go, lo, '122-8', '123-4', 'drawbar 3 with A.T.' )
    addParam( grp, go, lo, '123-5', '124-1', 'drawbar 3 with ctrlped' )

    addParam( grp, go, lo, '124-2', '124-5', 'drawbar 4' )
    addParam( grp, go, lo, '124-6', '125-2', 'drawbar 4 with wheel' )
    addParam( grp, go, lo, '125-3', '125-7', 'drawbar 4 with A.T.' )
    addParam( grp, go, lo, '125-8', '126-4', 'drawbar 4 with ctrlped' )

    addParam( grp, go, lo, '126-5', '126-8', 'drawbar 5' )
    addParam( grp, go, lo, '127-1', '127-5', 'drawbar 5 with wheel' )
    addParam( grp, go, lo, '127-6', '128-2', 'drawbar 5 with A.T.' )
    addParam( grp, go, lo, '128-3', '128-7', 'drawbar 5 with ctrlped' )

    addParam( grp, go, lo, '128-8', '129-3', 'drawbar 6' )
    addParam( grp, go, lo, '129-4', '129-8', 'drawbar 6 with wheel' )
    addParam( grp, go, lo, '130-1', '130-5', 'drawbar 6 with A.T.' )
    addParam( grp, go, lo, '130-6', '131-2', 'drawbar 6 with ctrlped' )

    addParam( grp, go, lo, '131-3', '131-6', 'drawbar 7' )
    addParam( grp, go, lo, '131-7', '132-3', 'drawbar 7 with wheel' )
    addParam( grp, go, lo, '132-4', '132-8', 'drawbar 7 with A.T.' )
    addParam( grp, go, lo, '133-1', '133-5', 'drawbar 7 with ctrlped' )

    addParam( grp, go, lo, '133-6', '134-1', 'drawbar 8' )
    addParam( grp, go, lo, '134-2', '134-6', 'drawbar 8 with wheel' )
    addParam( grp, go, lo, '134-7', '135-3', 'drawbar 8 with A.T.' )
    addParam( grp, go, lo, '135-4', '135-8', 'drawbar 8 with ctrlped' )

    addParam( grp, go, lo, '136-1', '136-4', 'drawbar 9' )
    addParam( grp, go, lo, '136-5', '137-1', 'drawbar 9 with wheel' )
    addParam( grp, go, lo, '137-2', '137-6', 'drawbar 9 with A.T.' )
    addParam( grp, go, lo, '137-7', '138-3', 'drawbar 9 with ctrlped' )

    addParam( grp, go, lo, '141-1', '     ', 'vib/chorus on/off' )
    addParam( grp, go, lo, '141-2', '     ', 'percussion on/off' )
    addParam( grp, go, lo, '141-3', '     ', 'perc harm 3rd on/off' )
    addParam( grp, go, lo, '141-4', '     ', 'perc decay fast on/off' )
    addParam( grp, go, lo, '141-5', '     ', 'perc vol soft on/off' )
    
    addParamsFX( grp, bitsOffsetFX['o'] )

# ================================================================
def makeMapPiano():

    grp = 'p'
    go = 0

    lo = -1
    addParam( grp, go, lo, '230-3', '     ', 'layer on/off' )
    addParam( grp, go, lo, '230-6', '     ', 'layer on/off (scene II)' )

    lo = 31
    addParam( grp, go, lo, '230-7', '231-5', 'volume' )
    addParam( grp, go, lo, '231-6', '232-5', 'volume change with wheel' )
    addParam( grp, go, lo, '232-6', '233-5', 'volume change with A.T.' )
    addParam( grp, go, lo, '233-6', '234-5', 'volume change with ctrlped' )

    lo = 32
#    addParam( grp, go, lo, '061-7', '065-6', 'preset ID' )

    lo = 12*8
    addParam( grp, go, lo, '243-1', '243-4', 'KB zones' )
    addParam( grp, go, lo, '243-5', '243-8', 'octave shift' )
    addParam( grp, go, lo, '244-1', '     ', 'pstick on/off' )
    addParam( grp, go, lo, '244-2', '     ', 'susped on/off' )
    addParam( grp, go, lo, '244-3', '244-5', 'piano type' )
    addParam( grp, go, lo, '244-6', '245-2', 'piano model slot' )
    addParam( grp, go, lo, '245-3', '245-4', 'piano model variation' )
    addParam( grp, go, lo, '245-5', '249-4', 'piano model ID/name' )
    addParam( grp, go, lo, '249-5', '     ', 'soft rel on/off' )
    addParam( grp, go, lo, '249-6', '     ', 'string res on/off' )
    addParam( grp, go, lo, '249-7', '     ', 'pedal noise on/off' )
    addParam( grp, go, lo, '249-8', '250-1', 'touch' )
    addParam( grp, go, lo, '250-2', '250-3', 'unison level' )
    addParam( grp, go, lo, '250-4', '250-5', 'dyn comp' )
    addParam( grp, go, lo, '250-7', '251-1', 'timbre' )

    addParamsFX( grp, bitsOffsetFX['p'] )

# ================================================================
def makeMapSynth():

    grp = 'y'
    go = 0

    lo = -1
    addParam( grp, go, lo, '377-5', '     ', 'layer on/off' )
    addParam( grp, go, lo, '378-2', '     ', 'layer on/off (scene II)' )

    lo = 31
    addParam( grp, go, lo, '378-3', '379-1', 'volume' )
    addParam( grp, go, lo, '379-2', '380-1', 'volume change with wheel' )
    addParam( grp, go, lo, '380-2', '381-1', 'volume change with A.T.' )
    addParam( grp, go, lo, '381-2', '382-1', 'volume change with ctrlped' )
    addParam( grp, go, lo, '389-8', '390-5', 'pan' )

    lo = 32
#    addParam( grp, go, lo, '071-5', '075-4', 'preset ID' )

    lo = 51*8
    addParam( grp, go, lo, '408-1', '     ', 'samples/analog' )
    addParam( grp, go, lo, '408-7', '410-2', 'sample slot' )
    addParam( grp, go, lo, '410-3', '414-2', 'sample ID/name' )

    lo = 48*8
    addParam( grp, go, lo, '575-6', '575-7', 'sample options' )
    addParam( grp, go, lo, '604-3', '     ', 'sample bright on/off' )

    lo = 51*8
    addParam( grp, go, lo, '414-3', '414-6', 'KB zones' )
    addParam( grp, go, lo, '414-7', '415-2', 'octave shift' )
    addParam( grp, go, lo, '415-3', '     ', 'pstick on/off' )
    addParam( grp, go, lo, '415-4', '415-7', 'pstick rng' )
    addParam( grp, go, lo, '415-8', '     ', 'susped on/off' )
    addParam( grp, go, lo, '416-1', '416-3', 'vibrato mode' )
    addParam( grp, go, lo, '453-5', '454-1', 'vibrato delay' )

    lo = 48*8
    addParam( grp, go, lo, '602-6', '603-4', 'vibrato rate' )
    addParam( grp, go, lo, '603-5', '604-2', 'vibrato amount' )

    lo = 51*8
    addParam( grp, go, lo, '416-7', '     ', 'mono on/off' )
    addParam( grp, go, lo, '416-6', '     ', 'legato on/off' )
    addParam( grp, go, lo, '416-8', '417-1', 'voice priority' )
    addParam( grp, go, lo, '417-2', '417-8', 'glide' )
    addParam( grp, go, lo, '423-7', '     ', 'KB hold' )
    addParam( grp, go, lo, '423-8', '     ', 'arpeggiator run on/off' )
    addParam( grp, go, lo, '424-1', '424-2', 'arpeggiator mode' )
    addParam( grp, go, lo, '424-3', '     ', 'arp pattern on/off' )
    addParam( grp, go, lo, '424-4', '     ', 'KB sync on/off' )
    addParam( grp, go, lo, '424-5', '425-3', 'arp range/env' )
    addParam( grp, go, lo, '425-4', '426-3', 'arp range/env with wheel' )
    addParam( grp, go, lo, '426-4', '427-3', 'arp range/env with A.T.' )
    addParam( grp, go, lo, '427-4', '428-3', 'arp range/env with ctrlped' )
    addParam( grp, go, lo, '428-4', '428-5', 'arp direction' )
    addParam( grp, go, lo, '428-6', '     ', 'arp zigzag on/off' )
    addParam( grp, go, lo, '428-7', '     ', 'arp MST CLK on/off' )
    addParam( grp, go, lo, '428-8', '429-6', 'arp rate/time' )
    addParam( grp, go, lo, '429-7', '430-6', 'arp rate/time with wheel' )
    addParam( grp, go, lo, '430-7', '431-6', 'arp rate/time with A.T.' )
    addParam( grp, go, lo, '431-7', '432-6', 'arp rate/time with ctrlped' )
    addParam( grp, go, lo, '432-7', '433-2', 'arp pattern length' )
    addParam( grp, go, lo, '433-3', '437-2', 'arpeggiator accent' )
    addParam( grp, go, lo, '437-3', '441-2', 'arpeggiator gate' )
    addParam( grp, go, lo, '441-3', '445-2', 'arpeggiator pan' )

    addParam( grp, go, lo, '445-3', '445-4', 'unison level' )
    addParam( grp, go, lo, '418-1', '     ', 'extern on/off' )
    addParam( grp, go, lo, '421-8', '422-6', 'extern program' )
    addParam( grp, go, lo, '445-5', '446-3', 'extern CC VAL1' )
    addParam( grp, go, lo, '446-4', '447-3', 'extern CC VAL1 with wheel' )
    addParam( grp, go, lo, '447-4', '448-3', 'extern CC VAL1 with A.T.' )
    addParam( grp, go, lo, '448-4', '449-3', 'extern CC VAL1 with ctrlped' )
    addParam( grp, go, lo, '449-4', '450-2', 'extern CC VAL2' )
    addParam( grp, go, lo, '450-3', '451-2', 'extern CC VAL2 with wheel' )
    addParam( grp, go, lo, '451-3', '452-2', 'extern CC VAL2 with A.T.' )
    addParam( grp, go, lo, '452-3', '453-2', 'extern CC VAL2 with ctrlped' )

    lo = 48*8
    addParam( grp, go, lo, '561-3', '561-4', 'analog type (knob 1)' )
    addParam( grp, go, lo, '561-8', '562-2', 'analog cat (knob 2)' )
    addParam( grp, go, lo, '562-4', '563-1', 'analog wave/partial (knob 3)' )
    addParam( grp, go, lo, '563-2', '563-8', 'osc ctrl' )
    addParam( grp, go, lo, '564-1', '564-8', 'osc ctrl with wheel' )
    addParam( grp, go, lo, '565-1', '565-8', 'osc ctrl with A.T.' )
    addParam( grp, go, lo, '566-1', '566-8', 'osc ctrl with ctrlped' )
    addParam( grp, go, lo, '567-1', '567-7', 'pitch fine' )
    addParam( grp, go, lo, '567-8', '568-5', 'pitch coarse' )
    addParam( grp, go, lo, '568-8', '569-6', 'osc env attack' )
    addParam( grp, go, lo, '569-7', '570-5', 'osc env decay' )
    addParam( grp, go, lo, '570-6', '571-4', 'osc env release' )
    addParam( grp, go, lo, '571-5', '572-3', 'osc env amt' )
    addParam( grp, go, lo, '572-4', '573-3', 'osc env amt with wheel' )
    addParam( grp, go, lo, '573-4', '574-3', 'osc env amt with A.T.' )
    addParam( grp, go, lo, '574-4', '575-3', 'osc env amt with ctrlped' )
    addParam( grp, go, lo, '575-4', '     ', 'osc env-to-pitch on/off' )
    addParam( grp, go, lo, '575-5', '     ', 'osc env velocity on/off' )
    addParam( grp, go, lo, '575-8', '576-1', 'LFO target' )
    addParam( grp, go, lo, '576-2', '576-4', 'LFO shape' )
    addParam( grp, go, lo, '576-5', '     ', 'LFO mst clk on/off' )
    addParam( grp, go, lo, '576-6', '577-4', 'LFO rate/time' )
    addParam( grp, go, lo, '577-5', '578-4', 'LFO rate/time with wheel' )
    addParam( grp, go, lo, '578-5', '579-4', 'LFO rate/time with A.T.' )
    addParam( grp, go, lo, '579-5', '580-4', 'LFO rate/time with ctrlped' )
    addParam( grp, go, lo, '580-5', '581-3', 'LFO mod amt' )
    addParam( grp, go, lo, '581-4', '582-3', 'LFO mod amt with wheel' )
    addParam( grp, go, lo, '582-4', '583-3', 'LFO mod amt with A.T.' )
    addParam( grp, go, lo, '583-4', '584-3', 'LFO mod amt with ctrlped' )
    addParam( grp, go, lo, '584-4', '585-2', 'amp env attack' )
    addParam( grp, go, lo, '585-3', '586-1', 'amp env decay' )
    addParam( grp, go, lo, '586-2', '586-8', 'amp env release' )
    addParam( grp, go, lo, '587-1', '587-2', 'amp env velocity' )
    addParam( grp, go, lo, '602-5', '     ', 'filter on/off' )
    addParam( grp, go, lo, '587-3', '587-5', 'filter type' )
    addParam( grp, go, lo, '587-6', '588-4', 'filter freq' )
    addParam( grp, go, lo, '588-5', '589-4', 'filter freq with wheel' )
    addParam( grp, go, lo, '589-5', '590-4', 'filter freq with A.T.' )
    addParam( grp, go, lo, '590-5', '591-4', 'filter freq with ctrlped' )
    addParam( grp, go, lo, '591-5', '592-3', 'filter resonance / freq HP' )
    addParam( grp, go, lo, '592-4', '593-3', 'filter res with wheel' )
    addParam( grp, go, lo, '593-4', '594-3', 'filter res with A.T.' )
    addParam( grp, go, lo, '594-4', '595-3', 'filter res with ctrlped' )
    addParam( grp, go, lo, '595-4', '595-5', 'filter track' )
    addParam( grp, go, lo, '595-6', '595-7', 'filter drive' )
    addParam( grp, go, lo, '595-8', '596-6', 'filter env amt' )
    addParam( grp, go, lo, '596-7', '597-6', 'filter env amt with wheel' )
    addParam( grp, go, lo, '597-7', '598-6', 'filter env amt with A.T.' )
    addParam( grp, go, lo, '598-7', '599-6', 'filter env amt with ctrlped' )
    addParam( grp, go, lo, '599-7', '600-5', 'filter env attack' )
    addParam( grp, go, lo, '600-6', '601-4', 'filter env decay' )
    addParam( grp, go, lo, '601-5', '602-3', 'filter env release' )
    addParam( grp, go, lo, '602-4', '     ', 'filter velocity on/off' )

    addParamsFX( grp, bitsOffsetFX['y'] )

# ================================================================
#  option to read bitmap files to augment/override ns4map's built-in map
#
#  this can be useful during the development process
#
def addMapsFromFile( group, filename, verbosity ):

    global paramdata
    global groupparams

    if not ospath.isfile( filename ):
        if verbosity >= 3:
            print( 'Warning: no file called', filename )
        return

    if verbosity >= 2:
        print( 'Reading', filename )
    lines = open( filename, 'r' ).readlines()

    numcolumns = len( lines[0].split(',') )
    numlayers = getNumLayers( group )

    if numcolumns != 2 * numlayers + 1:
        print( 'Error: wrong number of columns in', filename )
        exit()

    for n in range(1,len(lines)): # skip header
        if lines[n].strip() == '':
            # skip empty lines
            continue

        words = lines[n].split(',')
        if len(words) != numcolumns:
            print( 'Error reading', filename, ': every row should have the same number of entries separated by a comma' )
            exit()

        #  the loc of the first bit in layer A is used as the parameter's ID,
        #  so can't change that loc.  Can change everything else about the map, though.
        paramID = words[0].strip()
        name = words[-1].strip()

        #  replace this parameter's data if the parameter is already known, 
        #  and if it's not already known then add it to the dictionary
        newParam = False
        if not paramID in list(paramdata.keys()):
            print( 'Adding parameter:', name )
            newParam = True
            paramdata[paramID] = {}
            paramdata[paramID]['group'] = group
            paramdata[paramID]['numlayers'] = numlayers
            paramdata[paramID]['dependson'] = []
            groupparams[group].append( paramID )

        paramdata[paramID]['name'] = name

        # read one pair of columns for each layer.
        # the number of columns should be 1 more than the number of layers.
        for k in range(numlayers):

            locs = getLocs( words[2*k].strip(), words[2*k+1].strip(), 0 )

            if newParam:
                paramdata[paramID][k] = {}

            paramdata[paramID][k]['begloc'] = locs['begloc']
            paramdata[paramID][k]['endloc'] = locs['endloc']

# ================================================================
def checkForCommas( paramdata ):

    for paramID in paramdata.keys():
        name = paramdata[paramID]['name']
        if ',' in name:
            print( 'Error: commas in names are not allowed, and this one has a comma:', \
                    name )
            exit()

# ================================================================
#  get maps of where to find each parameter in the binary files
#
#  the data structures are documented at the top of this file
#
def getMaps( formatspath, verbosity ):

    global paramdata
    global groupparams

    paramdata = {}
    groupparams = {}
    for group in groups:
        groupparams[group] = []

    makeMapMaster()
    makeMapOrgan()
    makeMapPiano()
    makeMapSynth()

    if len(formatspath) > 0:
        if not ospath.isdir( formatspath ):
            print( 'Error: no directory called', formatspath )
            exit()
        for group in groups:
            addMapsFromFile( group, ospath.join( formatspath, mapFilenames[group] ), \
                             verbosity )

    checkForCommas( paramdata )

    return paramdata, groupparams

# ================================================================
#   useful during the development process
#
def writeMap( group, filename, paramdata, grouppparams, verbosity ):

    getMaps( '', verbosity )

    numlayers = getNumLayers( group )
    layerpad = '     '

    headers = []
    if numlayers == 1:
        headers.append( layerpad + ' beg  ' )
        headers.append( layerpad + ' end  ' )
    else:
        for k in range(numlayers):
            headers.append( ' lyr ' + ns4names.Letter[k] + ' beg ' )
            headers.append( ' lyr ' + ns4names.Letter[k] + ' end ' )
    headers.append( ' parameter' )

    if verbosity >= 2:
        print( 'Writing', filename )

    fid = open( filename, 'w' )
    fid.write( ','.join(headers) )
    fid.write( '\n' )
    for n in range(len(grouppparams[group])):
        row = []
        paramID = grouppparams[group][n]

        for k in range(numlayers):
            row.append( ' ' + makeLocStr( paramdata[paramID][k]['begloc'] ) + layerpad )
            row.append( ' ' + makeLocStr( paramdata[paramID][k]['endloc'] ) + layerpad )
        row.append( ' ' + paramdata[paramID]['name'] )

        fid.write( ','.join(row) )
        fid.write( '\n' )
    fid.close()

# ================================================================
#   useful during the development process
#
def writeMaps( outputDirectory, paramdata, grouppparams, verbosity ):

    if outputDirectory.strip() != '' and not ospath.isdir( outputDirectory ):
        print( 'Error: no directory called', outputDirectory )
        exit()

    for group in groups:
        filename = ospath.join( outputDirectory, mapFilenames[group] )
        writeMap( group, filename, paramdata, grouppparams, verbosity )

    #  tell the user where the output files were written
    if outputDirectory in ['.', '']:
        print( 'Outputs written to current working directory' )
    else:
        print( 'Outputs written to', outputDirectory )
    
# ================================================================
#  given a list of IDs for layer 1 FX in the synth section,
#  this appends the corresponding IDs for the organ and piano sections
#
def append_IDs_for_organ_and_piano_FX( listOfIDs ):

    for n in range(len(listOfIDs)):
        listOfIDs.append( shiftBit( listOfIDs[n], -530*8 ) ) # organ section
        listOfIDs.append( shiftBit( listOfIDs[n], -438*8 ) ) # piano section

    return listOfIDs

# ================================================================
#   shouldInterpretAs is a static dictionary used in ns4decode.py
#
shouldInterpretAs = {}

#  do the FX params first to exploit the automation in finishFX_IDx(), 
#  and then extend the lists to other params (if any)

shouldInterpretAs['FXmod1mode'] = ['713-1']
shouldInterpretAs['FXmod2mode'] = ['721-4']
shouldInterpretAs['EQlevel'] = ['722-1','723-7']
# EQmid, FXmod1rate, DelayTempo require special handling because 
# their interpretations depend on other parameters
shouldInterpretAs['EQmid'] = ['722-8']      
shouldInterpretAs['FXmod1rate'] = ['705-3']
shouldInterpretAs['DelayTempo'] = ['734-3']
shouldInterpretAs['EQfreq'] = ['724-6']
shouldInterpretAs['NonMorphableZeroToTen'] = ['733-1']
shouldInterpretAs['FXcompResponse'] = ['733-8']
shouldInterpretAs['FXdelayMode'] = ['745-5']
shouldInterpretAs['FXdelayFilterType'] = ['745-7']
shouldInterpretAs['FXdelayEffects'] = ['749-8']
shouldInterpretAs['FXdarkBright'] = ['754-4']
shouldInterpretAs['FXreverbType'] = ['754-6']
shouldInterpretAs['AmpEQmode'] = ['755-6']
shouldInterpretAs['MorphableZeroToTen'] = [ '709-2', '713-6', '717-5', \
                                            '728-5', '741-6', '746-1', '750-5']
shouldInterpretAs['OnOff'] = [ '705-1', '705-2', '713-5', '721-8', \
                               '732-8', '734-1', '734-2', '745-6', '750-4']

for key in shouldInterpretAs.keys():
    shouldInterpretAs[key] = append_IDs_for_organ_and_piano_FX( shouldInterpretAs[key] )

#  now extend to non-FX params

shouldInterpretAs['NonMorphableZeroToTen'].extend( ['417-2', '105-5'] )

shouldInterpretAs['MorphableZeroToTen'].extend( ['424-5', '563-2', '595-8'] )

shouldInterpretAs['OnOff'].extend( [ \
             '050-1', '050-2', '050-3', '050-4', '052-7', '054-8', '055-1', \
             '055-2', '084-5', '084-6', '084-7', '084-8', '085-1', \
             '085-2', '088-4', '095-3', '095-6', '103-5', '104-1', \
             '109-3', '114-1', '114-5', '141-1', '141-2', '141-4', '141-5', '141-3', \
             '230-3', '230-6', '244-1', '244-2', '249-5', '249-6', '249-7', \
             '377-5', '378-2', '401-6', '401-8', '415-3', '415-8', '416-7', '416-6', \
             '418-1', '423-8', '424-3', '424-4', '428-6', '428-7', '575-4', '575-5', \
             '576-5', '602-4', '602-5', '604-3'] )

shouldInterpretAs['IncludeExclude'] = ['423-7']

shouldInterpretAs['MorphableVolume'] = ['095-7', '230-7', '378-3']

shouldInterpretAs['MorphableExtern'] = ['445-5', '449-4']

shouldInterpretAs['Drawbar']  = ['117-1'] # first drawbar
for n in range(1,9):    # the other drawbars
    shiftAmount = (2*8+3)*n
    shouldInterpretAs['Drawbar'].append( shiftBit( shouldInterpretAs['Drawbar'][0], \
                                                   shiftAmount ) )
    # the drawbar interpretations depend on the organ model
    # in a way that is different for the 9th drawbar than for the other drawbars
    if n+1 == 9:
        shouldInterpretAs['Drawbar9'] = [shiftBit( shouldInterpretAs['Drawbar'][0], \
                                                   shiftAmount )]

shouldInterpretAs['ArpAccent'] = ['433-3']
shouldInterpretAs['ArpGate']   = ['437-3']
shouldInterpretAs['ArpPan']    = ['441-3']
shouldInterpretAs['ArpRate']   = ['428-8']
shouldInterpretAs['LFOrate']   = ['576-6']
shouldInterpretAs['LFOmodAmt'] = ['580-5']
shouldInterpretAs['OscEnvAmt'] = ['571-5']
shouldInterpretAs['FilterFreq'] = ['587-6']
shouldInterpretAs['FilterRes'] = ['591-5']
shouldInterpretAs['PstickRng'] = ['415-4']
shouldInterpretAs['KBzones'] = ['113-1', '243-1', '414-3']
shouldInterpretAs['fileVersion'] =['021-1', '048-1']
shouldInterpretAs['presetID'] =['056-5', '061-7', '071-5']
