#
#  -------------------------------------------------------------------------
#  MIT License
#  
#  Copyright (c) 2024 Randy
#  
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  -------------------------------------------------------------------------
#
#  this module is imported by ns4decode.py
#
#  this module provides the function getOptions()
#
from os import path as ospath
from os import name as osname
from sys import exit, argv
from pathlib import Path as pathlibPath

import ns4help
import ns4maps

# ================================================================
def showFilename( filename ):

    if filename == "":
        return '""'
    return filename

# ================================================================
def showOnOrOff( isOn ):

    if isOn:
        return( 'on' )
    return( 'off' )

# ================================================================
def showSettings( options ):

    if options['gui']:
        print( '    --gui'  )
    else:
        print( '    --nogui'  )
    if options['csv']:
        print( '    --csv'  )
    else:
        print( '    --nocsv'  )
    print( '    --writeto', showFilename( options['outputDirectory'] ) )
    print( '    --columnwidth', options['columnwidth'] )
    print( '    --file',        showFilename( options['chosenparamsFile'] ) )
    print( '    --find',       ','.join(options['find']) )
    print( '    --includeall',  showOnOrOff( options['includeall'] ) )
    print( '    --askfirst',    showOnOrOff( options['askfirst'] ) )
    print( '    --warnversion', showOnOrOff( options['warnversion'] ) )
    print( '    --verbosity', options['verbosity'] )
    if options['valueformat'] == 'decimal':
        print( '    --rawdecimal'  )
    if options['valueformat'] == 'hex':
        print( '    --rawhex' )
    if options['valueformat'] == 'binary':
        print( '    --rawbinary' )
    print( '    --use', showFilename( options['formatspath'] ) )
    if options['chosenparams']:
        print( 'Chosen parameters: ', options['chosenparams'] )

    # print( options )

# ================================================================
def getListOfIntegers( csvlist ):

    words = csvlist.split(',')
    intlist = []
    for n in range(len(words)):
        if not words[n].isnumeric():
            print( 'Error: parameter indexes must be integers, and this is not an integer:', words[n] )
            exit()
        intlist.append( int(words[n]) )

    return intlist

# ================================================================
def parseBoolean( word ):
    if word in ['no', 'No', 'NO', 'off', 'Off', '0', 'false', 'False', 'FALSE']:
        return False
    if word in ['yes', 'Yes', 'YES', 'on', 'On', '1', 'true', 'True', 'TRUE']:
        return True
    print( 'Error:', word, 'is not a valid boolean value' )
    exit()

# ================================================================
def needsValue( words, idx, needWhat ):
    if idx+1 >= len(words):
        print( 'Error:', needWhat, 'not specified' )
        exit()

# ================================================================
def parsePath( words, idx ):

    if idx > 0:
        # reading from command line
        result = words[idx+1]
    else:
        # if reading from preferences file, then don't break at whitespace
        result = ' '.join( words[idx+1:] )

        # https://stackoverflow.com/questions/1854/how-to-identify-which-os-python-is-running-on
        if osname != 'nt':
            # If not running on Windows, then remove backslashes (if any).
            # On Windows, need to keep the backslashes (used as directory separators).
            result = result.replace('\\', '')

    # need to remove quotes (if any) because a filename will be appended
    result = result.replace( '"', '' )

    return result

# ================================================================
def parseOption( options, words, idx ):

    foundOption = False

    if words[idx] in ['-g','--gui']:
        options['gui'] = True
        idx += 1
        foundOption = True
    elif words[idx] in ['--nogui']:
        options['gui'] = False
        idx += 1
        foundOption = True
    elif words[idx] in ['--csv']:
        options['csv'] = True
        idx += 1
        foundOption = True
    elif words[idx] in ['--nocsv']:
        options['csv'] = False
        idx += 1
        foundOption = True
    elif words[idx] in ['--rawbinary']:
        options['valueformat'] = 'binary'
        idx += 1
        foundOption = True
    elif words[idx] in ['--rawdecimal']:
        options['valueformat'] = 'decimal'
        idx += 1
        foundOption = True
    elif words[idx] in ['--rawhex']:
        options['valueformat'] = 'hex'
        idx += 1
        foundOption = True
    elif words[idx] in ['-x', '--exclude']:
        options['includeall'] = False
        idx += 1
        foundOption = True
    elif words[idx] in ['--includeall']:
        needsValue( words, idx, 'on/off' )
        options['includeall'] = parseBoolean(words[idx+1])
        idx += 2
        foundOption = True
    elif words[idx] in ['--warnversion']:
        needsValue( words, idx, 'on/off' )
        options['warnversion'] = parseBoolean(words[idx+1])
        idx += 2
        foundOption = True
    elif words[idx] in ['--askfirst']:
        needsValue( words, idx, 'on/off' )
        options['askfirst'] = parseBoolean(words[idx+1])
        idx += 2
        foundOption = True
    elif words[idx] in ['-c', '--columnwidth']:
        needsValue( words, idx, 'column width' )
        options['columnwidth'] = int(words[idx+1])
        idx += 2
        foundOption = True
    elif words[idx] in ['-v', '--verbosity']:
        needsValue( words, idx, 'verbosity' )
        options['verbosity'] = int(words[idx+1])
        if options['verbosity'] < 0 or options['verbosity'] > 3:
            print( 'Error: verbosity must be an integer from 0 to 3' )
            exit() 
        idx += 2
        foundOption = True
    elif words[idx] in ['-w', '--writeto']:
        needsValue( words, idx, 'output folder' )
        options['outputDirectory'] = parsePath( words, idx )
        if not options['outputDirectory'].strip() == "" and \
           not ospath.isdir( options['outputDirectory'] ):
            print( 'Error: no folder called', options['outputDirectory'] )
            exit()
        idx += 2
        foundOption = True
    elif words[idx] in ['-f', '--file']:
        needsValue( words, idx, 'filename' )
        chosenparamsFile = parsePath( words, idx )
        if not chosenparamsFile.strip() == "" and ospath.isfile( chosenparamsFile ):
            print( 'Warning:', chosenparamsFile, 'already exists' )
            yesorno = input( 'Do you want to replace it (y/n)? ' )
            if yesorno.strip() not in ['y', 'yes', 'Y', 'Yes', 'YES']:
                print( 'Exiting without writing anything.' )
                exit()
        if not chosenparamsFile.strip() == "" and ospath.isdir( chosenparamsFile ):
            print( 'Error:', chosenparamsFile, 'is a directory (folder)' )
            exit()
        options['chosenparamsFile'] = chosenparamsFile
        idx += 2
        foundOption = True
    elif words[idx] in ['--use']:
        needsValue( words, idx, 'folder' )
        options['formatspath'] = words[idx+1]
        idx += 2
        foundOption = True
    elif words[idx] in ['--layer', '--layers']:
        needsValue( words, idx, 'layer' )
        options['chosenlayers'] = words[idx+1]
        idx += 2
        foundOption = True
    elif words[idx] in ['--find']:
        needsValue( words, idx, 'find' )
        options['find'] = words[idx+1].split(',')
        idx += 2
        foundOption = True
    else:
        for group in ns4maps.groups:
            if words[idx] == '-' + group:
                if not group in options['chosenparams'].keys():
                    # initialize
                    options['chosenparams'][group] = []
                needsValue( words, idx, 'parameter index' )
                options['chosenparams'][group].extend( getListOfIntegers( words[idx+1] ) )
                idx += 2
                foundOption = True
                break

    return options, idx, foundOption

# ================================================================
def readOptionsFromFile():

    options = {}
    #  the 'formatspath' option is used to make ns4decode read bitmaps/names from files,
    #  to augment/replace it's built-in bitmaps/names
    options['formatspath'] = '' 
    options['chosenparams'] = {} # list of user-selected parameters to view
    options['chosenparamsFile'] = '' # blank means write to the screen
    options['chosenlayers'] = ''  # list of user-selected layers to view
    options['outputDirectory'] = '' # blank means write to the input directory
    options['csv'] = True #  if True, then write output to CSV files
    options['gui'] = False #  if True, then show output in graphic form
    options['includeall'] = True #  if False, then parameters are only shown when their section/layer/grouping is active
    options['askfirst'] = True
    options['verbosity'] = 1
    options['warnversion'] = True
    options['valueformat'] = ''   # default is to show the interpreted value
    options['find'] = ['ns4p','ns4l','ns4o','ns4n','ns4y'] # only decode these file types in whole-folder mode

    # 0 is default because (1) it's not arbitrary, 
    #                      (2) it doesn't interfere with a spreadsheet-app's formatting
    options['columnwidth']  =  0

    #whereinstalled = ospath.dirname(ospath.realpath(__file__))
    # https://stackoverflow.com/questions/2632199
    whereinstalled = ospath.abspath(ospath.dirname(argv[0]))
    preferencesfile = ospath.join( whereinstalled, 'preferences.txt' )
    if ospath.isfile( preferencesfile ):
        if options['verbosity'] >= 2:
            print( 'Reading', preferencesfile )
        for line in open( preferencesfile, 'r' ).readlines():
            if line.strip().startswith( '#' ):
                continue
            words = line.split()
            if len(words) > 0:
                options, dummy, foundOption = parseOption( options, words, 0 )

    return options

# ================================================================
#  parse command-line options
# 
def getOptionsFromCommandLine( options ):

    if len(argv) < 2:
        ns4help.showIntro()
        exit()

    options['listparams'] = False
    options['listparamsnocodes'] = False
    options['writenames'] = False
    options['writemaps'] = False
    options['inputFilenames'] = []

    shouldDecodeSomething = True

    a = 1
    while a < len(argv):
        # ----------------------------------
        #  options that show information without decoding files
        # ----------------------------------
        if argv[a] in ['-h', '--help', '-H', '--Help', '--HELP']:
            ns4help.showHelp()
            exit()
        if argv[a].startswith( '--help' ):
            # this catches attempts to use the help options in previous versions of ns4decode
            ns4help.showHelp()
            exit()
        if argv[a] in ['--settings']:
            showSettings( options )
            exit()
        if argv[a] in ['--version']:
            ns4help.showVersion()
            exit()
        if argv[a] in ['-l', '--listparams']:
            options['listparams'] = True
            a += 1
            shouldDecodeSomething = False
            break
        if argv[a] in ['--listparamsnocodes']:
            options['listparamsnocodes'] = True
            a += 1
            shouldDecodeSomething = False
            break
        if argv[a] in ['--names']:
            options['writenames'] = True
            a += 1
            shouldDecodeSomething = False
            continue  # in case --bitmaps is also specified
        if argv[a] in ['--bitmaps']:
            options['writemaps'] = True
            a += 1
            shouldDecodeSomething = False
            continue  # in case --names is also specified
        # ----------------------------------
        #  options that customize behavior when decoding files
        # ----------------------------------
        options, a, foundOption = parseOption( options, argv, a )
        if foundOption:
            continue
        else:
            options['inputFilenames'] = argv[a:]
            break

    if options['verbosity'] > 2:
        showSettings( options )
    if shouldDecodeSomething and len(options['inputFilenames']) < 1:
        print( 'Warning: no ns4 files specified' )

    return options

# ================================================================
def getInputFilenames( folderName, searchString ):

    inputFilenames = []
    files = pathlibPath( folderName ).rglob( searchString )
    for f in files:
        inputFilenames.append( ospath.abspath(f) )
    return inputFilenames

# ================================================================
#  If one of the input files is actually a folder, then 
#  find all of the ns4 files it contains (recursively)
#  and insert them into the list-of-files in place of the folder
#
def unpackFolders( options ):

    anyFolders = False
    inputFilenames = []
    for n in range(len(options['inputFilenames'])):
        name = options['inputFilenames'][n]
        if ospath.isdir( name ):
            #  name is a directory (folder), so find all of the ns4 files under it
            for filetype in options['find']:
                inputFilenames.extend( getInputFilenames( name, '*.' + filetype ) )
            anyFolders = True
        else:
            #  name is a single ns4 file, so keep it
            inputFilenames.append( name )

    if anyFolders:
        options['inputFilenames'] = inputFilenames
        if options['askfirst'] and len(options['chosenparams'].keys()) == 0:
            print( 'Found', len(inputFilenames), 'ns4 files.' )
            yesorno = input( 'Do you want to decode them all (y/n)? ' )
            if not yesorno.strip() in ['y', 'yes', 'Y', 'Yes', 'YES']:
                print( 'Exiting without decoding any files.' )
                exit()

    return options

# ================================================================
def getOptions():

    options = readOptionsFromFile()
    options = getOptionsFromCommandLine( options )

    if len(options['chosenparams'].keys()) > 0:
        # if output will be written to stdout,
        # then suppress file I/O messages
        options['verbosity'] = 0
    
    if 0:
        # confirm that command-line options were parsed correctly (for debugging)
        print( options )
    
    options = unpackFolders( options )

    return options
